"""
Functions for visualizing system load over time in a Unicode graph.

The one you probably want to call is get_load_string().
"""

import os

from . import px_cpuinfo
from . import px_terminal

from typing import Tuple, Optional


physical, logical = px_cpuinfo.get_core_count()
physical_string = px_terminal.bold(str(physical) + " cores")
cores_string = f"[{physical_string} | {logical} virtual]"


def average_to_level(average, peak):
    level = 3 * (average / peak)
    return int(round(level))


def averages_to_levels(avg0, avg1, avg2):
    """
    Converts three load averages into three levels.

    A level is a 0-3 integer value.

    This function returns the three leves, plus the peak value the levels are
    based on.
    """
    peak = max(avg0, avg1, avg2)
    if peak < 1.0:
        peak = 1.0

    l0 = average_to_level(avg0, peak)
    l1 = average_to_level(avg1, peak)
    l2 = average_to_level(avg2, peak)
    return (l0, l1, l2, peak)


def levels_to_graph(levels):
    """
    Convert an array of levels into a unicode string graph.

    Each level in the levels array is an integer 0-3. Those levels will be
    represented in the graph by 1-4 dots each.

    The returned string will contain two levels per rune.
    """
    if len(levels) % 2 == 1:
        # Left pad uneven-length arrays with an empty column
        levels = [-1] + levels

    # https://en.wikipedia.org/wiki/Braille_Patterns#Identifying.2C_naming_and_ordering
    LEFT_BAR = [0x00, 0x40, 0x44, 0x46, 0x47]
    RIGHT_BAR = [0x00, 0x80, 0xA0, 0xB0, 0xB8]

    graph = ""
    for index in range(0, len(levels) - 1, 2):
        left_level = levels[index] + 1
        right_level = levels[index + 1] + 1
        code = 0x2800 + LEFT_BAR[left_level] + RIGHT_BAR[right_level]
        graph += chr(code)

    return graph


def get_load_values() -> Tuple[float, float, float]:
    """
    Returns three system load numbers:
    * The first is the average system load over the last 0m-1m
    * The second is the average system load over the last 1m-5m
    * The third is the average system load over the last 5m-15m
    """
    avg1, avg5, avg15 = os.getloadavg()

    avg0to1 = avg1
    avg1to5 = (5 * avg5 - avg1) / 4.0
    avg5to15 = (15 * avg15 - 5 * avg5) / 10.0

    return (avg0to1, avg1to5, avg5to15)


def get_load_string(load_values: Optional[Tuple[float, float, float]] = None) -> str:
    """
    Example return string, underlines indicate bold:
    "1.5  [4 cores | 8 virtual]  [15m history: GRAPH]"
     ^^^   ^^^^^^^                                  ^^^^^

    Load number is color coded:
    * <= physical core count: Green
    * <= virtual core count: Yellow
    * Larger: Red
    """
    if load_values is None:
        load_values = get_load_values()

    avg0to1, avg1to5, avg5to15 = load_values

    load_string = f"{avg0to1:.1f}"
    if avg0to1 <= physical:
        load_string = px_terminal.green(load_string)
    elif avg0to1 <= logical:
        load_string = px_terminal.yellow(load_string)
    else:
        load_string = px_terminal.red(load_string)

    recent, between, old, _ = averages_to_levels(avg0to1, avg1to5, avg5to15)
    graph = levels_to_graph([old] * 10 + [between] * 4 + [recent])

    # Increase intensity for more recent times
    graph = px_terminal.faint(graph[0:3]) + graph[3:6] + px_terminal.bold(graph[6:])

    return f"{load_string}  {cores_string}  [15m history: {graph}]"



import sys
import errno
import getpass
import datetime
import operator

import os
from . import px_file
from . import px_process
from . import px_ipc_map
from . import px_terminal
from . import px_cwdfriends
from . import px_loginhistory


from typing import MutableSet
from typing import Optional
from typing import Iterable
from typing import List
from typing import Tuple


def println(fd: int, string: str) -> None:
    os.write(fd, string.encode() + b"\n")


def find_process_by_pid(
    pid: int, processes: List[px_process.PxProcess]
) -> Optional[px_process.PxProcess]:
    for process in processes:
        if process.pid == pid:
            return process

    return None


def print_command_line(fd: int, process: px_process.PxProcess) -> None:
    """
    Print command line separated by linefeeds rather than space, this adds
    readability to long command lines
    """
    array = process.get_command_line_array()
    println(fd, array[0])
    for parameter in array[1:]:
        if parameter == "":
            # Print empty parameters as "", otherwise the printout just looks
            # broken.
            parameter = '""'
        println(fd, "  " + parameter)


def print_process_subtree(
    fd: int,
    process: px_process.PxProcess,
    indentation: int,
    lines: List[Tuple[str, px_process.PxProcess]],
) -> None:
    lines.append(("  " * indentation + str(process), process))
    for child in sorted(
        process.children, key=operator.attrgetter("lowercase_command", "pid")
    ):
        print_process_subtree(fd, child, indentation + 1, lines)


def print_process_tree(fd: int, process: px_process.PxProcess) -> None:
    # Contains tuples; the line to print and the process that line is for
    lines_and_processes: List[Tuple[str, px_process.PxProcess]] = []

    # List all parents up to the top
    parents = []
    here = process
    while here.parent is not None:
        # FIXME: There can actually be loops; we must detect that hand handle it appropriately
        parents.append(here.parent)
        here = here.parent

    # Print all parents
    indentation = 0
    for parent in reversed(parents):
        lines_and_processes.append(("  " * indentation + str(parent), parent))
        indentation += 1

    # Print ourselves
    bold_process = str(px_terminal.bold(str(process)))
    lines_and_processes.append(
        ("--" * (indentation - 1) + "> " + bold_process, process)
    )
    indentation += 1

    # Print all our child trees
    for child in sorted(
        process.children, key=operator.attrgetter("lowercase_command", "pid")
    ):
        print_process_subtree(fd, child, indentation, lines_and_processes)

    # Add an owners column to the right of the tree
    tree_width = max(
        map(lambda lp: px_terminal.visual_length(lp[0]), lines_and_processes)
    )
    lines = []
    current_user = os.environ.get("SUDO_USER") or getpass.getuser()
    for line_and_process in lines_and_processes:
        line = line_and_process[0]

        # NOTE: This logic should match its friend in
        # px_terminal.py/to_screen_lines()
        owner = line_and_process[1].username
        if owner == "root":
            owner = px_terminal.faint(owner)
        elif owner != current_user:
            # Neither root nor ourselves, highlight!
            owner = px_terminal.bold(owner)

        if line_and_process[1].pid == process.pid:
            sudo_user = process.get_sudo_user()
            if sudo_user:
                owner += ", $SUDO_USER=" + sudo_user

        padding = " " * (tree_width + 2 - px_terminal.visual_length(line))
        lines.append(line + padding + owner)

    println(fd, "\n".join(lines))


def to_relative_start_string(base, relative):
    delta_seconds = base.age_seconds - relative.age_seconds
    delta_string = px_process.seconds_to_str(abs(delta_seconds))
    before_or_after = "before"
    if delta_seconds > 0:
        before_or_after = "after"
    elif delta_seconds == 0:
        delta_string = "just"
        if relative.pid > base.pid:
            before_or_after = "after"
    return f"{relative} was started {delta_string} {before_or_after} {base}"


def get_closest_starts(
    process: px_process.PxProcess, all_processes: List[px_process.PxProcess]
) -> List[px_process.PxProcess]:
    """
    Return the processes that were started closest in time to the base process.

    All processes started within 1s of the base process are returned, or the
    five closest if not at least five were that close.
    """
    by_temporal_vicinity = sorted(
        all_processes, key=lambda p: abs(p.age_seconds - process.age_seconds)
    )

    closest_raw = set()
    for close in by_temporal_vicinity:
        delta_seconds = abs(close.age_seconds - process.age_seconds)

        if delta_seconds <= 1:
            closest_raw.add(close)
            continue

        # "5" is arbitrarily chosen, look at the printouts to see if it needs tuning
        if len(closest_raw) > 5:
            break

        closest_raw.add(close)

    # Remove ourselves from the closest processes list
    closest: Iterable[px_process.PxProcess] = filter(
        lambda p: p is not process, closest_raw
    )

    # Sort closest processes by age, command and PID in that order
    closest = sorted(closest, key=operator.attrgetter("command", "pid"))
    closest = sorted(closest, key=operator.attrgetter("age_seconds"), reverse=True)
    return closest


def print_processes_started_at_the_same_time(fd, process, all_processes):
    println(fd, "Other processes started close to " + str(process) + ":")
    for close in get_closest_starts(process, all_processes):
        println(fd, "  " + to_relative_start_string(process, close))


def print_users_when_process_started(fd: int, process: px_process.PxProcess) -> None:
    println(fd, "Users logged in when " + str(process) + " started:")
    users = px_loginhistory.get_users_at(process.start_time)
    if not users:
        println(
            fd,
            "  <Nobody found, either nobody was logged in or the wtmp logs have been rotated>",
        )
        return

    for user in sorted(users):
        println(fd, "  " + user)


def to_ipc_lines(ipc_map: px_ipc_map.IpcMap) -> Iterable[str]:
    return_me = []
    for target in sorted(ipc_map.keys(), key=operator.attrgetter("name", "pid")):
        channels = ipc_map[target]
        channel_names: MutableSet[str] = set()
        for channel_name in map(str, channels):
            channel_names.add(channel_name)
        for channel_name in sorted(channel_names):
            return_me.append(f"{px_terminal.bold(str(target))}: {channel_name}")

    return return_me


def print_cwd_friends(fd, process, all_processes, all_files):
    friends = px_cwdfriends.PxCwdFriends(process, all_processes, all_files)

    cwd_suffix = ""
    if friends.cwd:
        cwd_suffix = " (" + px_terminal.bold(friends.cwd) + ")"
    println(fd, "Others sharing this process' working directory" + cwd_suffix)
    if not friends.cwd:
        sudo_px = px_terminal.bold(f"sudo px {process.pid}")
        println(fd, f'  Working directory unknown, try again or try "{sudo_px}"')
        return

    if friends.cwd == "/":
        println(fd, "  Working directory too common, never mind.")
        return

    if len(friends.friends) == 0:
        println(fd, "  Nobody else shares this working directory.")
        return

    for friend in friends.friends:
        println(fd, "  " + str(friend))


def print_fds(
    fd: int, process: px_process.PxProcess, processes: Iterable[px_process.PxProcess]
) -> None:
    # It's true, I measured it myself /johan.walles@gmail.com
    println(
        fd,
        datetime.datetime.now().isoformat()
        + ": Now invoking lsof, this can take over a minute on a big system...",
    )

    # Flush what we have so far so the user has something to read during the pause.
    # This is useful when piping output into a pager like moar or less.

    # NOTE: If we switch to writing to file-like objects we should flush here,
    # our println() function flushes implicitly.

    files = px_file.get_all()
    println(fd, datetime.datetime.now().isoformat() + ": lsof done, proceeding.")

    println(fd, "")
    print_cwd_friends(fd, process, processes, files)

    is_root = os.geteuid() == 0
    ipc_map = px_ipc_map.IpcMap(process, files, processes, is_root=is_root)

    println(fd, "")
    println(fd, "File descriptors:")
    println(fd, "  stdin : " + ipc_map.fds[0])
    println(fd, "  stdout: " + ipc_map.fds[1])
    println(fd, "  stderr: " + ipc_map.fds[2])
    # Note that we used to list all FDs here, but some processes (like Chrome)
    # has silly amounts, making the px output unreadable. Users should consult
    # lsof directly for the full list.

    println(fd, "")
    println(fd, "Network connections:")
    # FIXME: Print "nothing found" or something if we don't find anything to put
    # here, maybe with a hint to run as root if we think that would help.
    for connection in sorted(
        ipc_map.network_connections, key=operator.attrgetter("name")
    ):
        println(fd, "  " + str(connection))

    println(fd, "")
    println(fd, "Inter Process Communication:")
    for line in to_ipc_lines(ipc_map):
        println(fd, "  " + line)

    println(fd, "")
    lsof = px_terminal.bold(f"sudo lsof -p {process.pid}")
    watch_lsof = px_terminal.bold(f"sudo watch lsof -p {process.pid}")
    println(
        fd,
        f'For a list of all open files, do "{lsof}", '
        f'or "{watch_lsof}" for a live view.',
    )

    if os.getuid() != 0:
        println(fd, "")
        println(
            fd,
            "NOTE: This information might be incomplete, "
            "running as root sometimes produces better results.",
        )


def print_start_time(fd: int, process: px_process.PxProcess) -> None:
    println(
        fd,
        "{} {} was started by {}, at {}.".format(
            px_terminal.bold(process.age_s + " ago"),
            process.command,
            px_terminal.bold(process.username),
            px_terminal.bold(process.start_time.isoformat()),
        ),
    )

    if process.cpu_time_seconds and process.age_seconds:
        cpu_percent = 100.0 * process.cpu_time_seconds / process.age_seconds
        println(
            fd,
            "{} has been its average CPU usage since then, or {}/{}".format(
                px_terminal.bold(f"{cpu_percent:.1f}%"),
                px_terminal.bold(process.cpu_time_s),
                px_terminal.bold(process.age_s),
            ),
        )


def print_pid_info(fd: int, pid: int) -> None:
    processes = px_process.get_all()

    process = find_process_by_pid(pid, processes)
    if not process:
        sys.exit(f"No such PID: {pid}")

    print_process_info(fd, process, processes)


def print_process_info(
    fd: int, process: px_process.PxProcess, processes: List[px_process.PxProcess]
) -> None:
    print_command_line(fd, process)

    # Print a process tree with all PID's parents and all its children
    println(fd, "")
    print_process_tree(fd, process)

    println(fd, "")
    print_start_time(fd, process)

    println(fd, "")
    print_processes_started_at_the_same_time(fd, process, processes)

    println(fd, "")
    print_users_when_process_started(fd, process)

    # List all files PID has open
    println(fd, "")
    try:
        print_fds(fd, process, processes)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise
        println(
            fd,
            'Can\'t list IPC / network sockets, make sure "lsof" is installed and in your $PATH',
        )



#!/usr/bin/python

"""px - ps and top for Human Beings
     https://github.com/walles/px

Usage:
  px [--debug] [--sort=cpupercent] [--no-username] [filter string]
  px [--debug] [--no-pager] [--color] <PID>
  px [--debug] --top [filter string]
  px [--debug] --tree [filter string]
  px --install
  px --help
  px --version

In the base case, px list all processes much like ps, but with the most
interesting processes last. A process is considered interesting if it has high
memory usage, has used lots of CPU or has been started recently.

If the optional filter string parameter is specified, processes will be shown if:
* The filter matches the user name of the process
* The filter matches a substring of the command line

If the optional PID parameter is specified, you'll get detailed information
about that particular PID.

In --top mode, a new process list is shown every second. The most CPU heavy
processes are on top. In this mode, CPU times are counted from when you first
invoked px, rather than from when each process started. This gives you a picture
of which processes are most active right now.

In --tree mode, a process tree is shown. Filtering is supported, try
"px --tree firefox" for example.

--top: Show a continuously refreshed process list
--tree: Print a process tree
--debug: Print debug logs (if any) after running
--install: Install px, ptop and pxtree in /usr/local/bin/
--no-pager: Print PID info to stdout rather than to a pager
--sort=cpupercent: Order processes by CPU percentage only
--no-username: Don't show the username column in px output
--color: Force color output even when piping
--help: Print this help
--version: Print version information
"""

import operator
import platform
import logging
import sys
import io
import os

from . import px_pager
from . import px_install
from . import px_process
from . import px_terminal
from . import px_processinfo

from typing import Optional, List


ERROR_REPORTING_HEADER = """
---

Problems detected, please send this text to one of:
* https://github.com/walles/px/issues/new
* johan.walles@gmail.com
"""


def install(argv: List[str]) -> None:
    """Find full path to self"""
    if not argv:
        sys.stderr.write("ERROR: Can't find myself, can't install\n")
        return

    px_pex = argv[0]
    if not px_pex.endswith(".pex"):
        sys.stderr.write("ERROR: Not running from .pex file, can't install\n")
        return

    px_install.install(px_pex, "/usr/local/bin/px")
    px_install.install(px_pex, "/usr/local/bin/ptop")
    px_install.install(px_pex, "/usr/local/bin/pxtree")


# This is the setup.py entry point
def main():
    argv = list(sys.argv)

    loglevel = logging.ERROR
    while "--debug" in argv:
        argv.remove("--debug")
        loglevel = logging.DEBUG

    stringIO = io.StringIO()
    configureLogging(loglevel, stringIO)

    try:
        _main(argv)
    except Exception:
        LOG = logging.getLogger(__name__)
        LOG.exception("Uncaught Exception")

    handleLogMessages(stringIO.getvalue())


# This method inspired by: https://stackoverflow.com/a/9534960/473672
def configureLogging(loglevel: int, stringIO: io.StringIO) -> None:
    rootLogger = logging.getLogger()
    rootLogger.setLevel(loglevel)

    handlers = []
    for handler in rootLogger.handlers:
        handlers.append(handler)
    for handler in handlers:
        rootLogger.removeHandler(handler)

    handler = logging.StreamHandler(stringIO)
    formatter = logging.Formatter(
        fmt="%(asctime)s %(levelname)-8s %(message)s", datefmt="%Y-%m-%d %H:%M:%S%Z"
    )
    handler.setFormatter(formatter)

    rootLogger.addHandler(handler)


def handleLogMessages(messages: Optional[str]) -> None:
    if not messages:
        return

    sys.stderr.write(ERROR_REPORTING_HEADER)
    sys.stderr.write("\n")

    # If this fails, run "tox.sh" once and the "version.py" file will be created
    # for you.
    #
    # NOTE: If we "import version" at the top of this file, we will depend on it
    # even if we don't use it. And this will make test avoidance fail to avoid
    # px.py tests every time you make a new commit (because committing recreates
    # version.py).
    from . import version

    sys.stderr.write("px version: " + version.VERSION + "\n")

    sys.stderr.write("\n")
    sys.stderr.write("Python version: " + sys.version + "\n")
    sys.stderr.write("\n")
    sys.stderr.write("Platform info: " + platform.platform() + "\n")
    sys.stderr.write("\n")
    sys.stderr.write(messages)
    sys.stderr.write("\n")
    sys.exit(1)


def _main(argv: List[str]) -> None:
    if "--install" in argv:
        install(argv)
        return

    if "--help" in argv:
        print(__doc__)
        return

    if "--version" in argv:
        # If this fails, run "tox.sh" once and the "version.py" file will be created for you.
        #
        # NOTE: If we "import version" at the top of this file, we will depend on it even if
        # we don't use it. And this will make test avoidance fail to avoid px.py tests every
        # time you make a new commit (because committing recreates version.py).
        from . import version

        print(version.VERSION)
        return

    with_pager: Optional[bool] = None
    with_color: Optional[bool] = None
    with_username = True
    top: bool = False
    tree: bool = False
    sort_cpupercent: bool = False

    while "--no-pager" in argv:
        with_pager = False
        argv.remove("--no-pager")
    if with_pager is None:
        with_pager = sys.stdout.isatty()

    while "--color" in argv:
        with_color = True
        argv.remove("--color")
    if with_color is None:
        with_color = sys.stdout.isatty()
    if not with_color:
        px_terminal.disable_color()

    while "--top" in argv:
        top = True
        argv.remove("--top")
    if os.path.basename(argv[0]).endswith("top"):
        top = True

    while "--tree" in argv:
        tree = True
        argv.remove("--tree")
    if os.path.basename(argv[0]).endswith("tree"):
        tree = True

    while "--sort=cpupercent" in argv:
        sort_cpupercent = True
        argv.remove("--sort=cpupercent")

    while "--no-username" in argv:
        # Ref: https://github.com/walles/px/issues/88#issuecomment-945099485
        with_username = False
        argv.remove("--no-username")

    if len(argv) > 2:
        sys.stderr.write("ERROR: Expected zero or one argument but got more\n\n")
        print(__doc__, file=sys.stderr)
        sys.exit(1)

    search = ""
    if len(argv) == 2:
        if argv[1].startswith("--"):
            sys.stderr.write(f"ERROR: Unknown argument: {argv[1]}\n\n")
            print(__doc__, file=sys.stderr)
            sys.exit(1)

        search = argv[1]

    if top and tree:
        sys.stderr.write("ERROR: --top and --tree are mutually exclusive\n\n")
        print(__doc__, file=sys.stderr)
        sys.exit(1)

    if top:
        # Pulling px_top in on demand like this improves test result caching
        from . import px_top

        px_top.top(search=search)
        return

    if tree:
        # Pulling px_tree in on demand like this improves test result caching
        from . import px_tree

        px_tree.tree(search=search)
        return

    try:
        pid = int(search)
        if not with_pager:
            px_processinfo.print_pid_info(sys.stdout.fileno(), pid)
            return

        # Page it!
        processes = px_process.get_all()
        process = px_processinfo.find_process_by_pid(pid, processes)
        if not process:
            sys.exit(f"No such PID: {pid}")

        px_pager.page_process_info(process, processes)
        return
    except ValueError:
        # It's a search filter and not a PID, keep moving
        pass

    procs = list(filter(lambda p: p.match(search), px_process.get_all()))

    columns: Optional[int] = None
    try:
        _, columns = px_terminal.get_window_size()
    except px_terminal.TerminalError:
        columns = None

    # Print the most interesting processes last; there are lots of processes and
    # the end of the list is where your eyes will be when you get the prompt back.
    procs = px_process.order_best_last(procs)
    if sort_cpupercent:
        procs = list(filter(lambda p: p.cpu_percent is not None, procs))
        procs = sorted(procs, key=operator.attrgetter("cpu_percent"))
    if search:
        # Put exact search matches last. Useful for "px cat" or other short
        # search strings with tons of hits.
        procs = sorted(procs, key=lambda p: p.command == search)
    lines = px_terminal.to_screen_lines(procs, None, None, with_username)

    if columns:
        for line in lines:
            print(px_terminal.crop_ansi_string_at_length(line, columns))
    else:
        try:
            print("\n".join(lines))
        except BrokenPipeError:
            # This happens when piping to a command that exits before we finish
            pass


if __name__ == "__main__":
    main()



"""
Functions for visualizing where IO is bottlenecking.
"""

import datetime
import math
import re
import os

from . import px_units
from . import px_terminal
from . import px_exec_util

from typing import List
from typing import Dict
from typing import Tuple
from typing import Optional


# Matches output lines in "netstat -ib" on macOS.
#
# Extracted columns are interface name, incoming bytes count and outgoing bytes
# count.
#
# If you look carefully at the output, this regex will only match lines with
# error counts, which is only one line per interface.
NETSTAT_IB_LINE_RE = re.compile(
    r"^([^ ]+).*[0-9]+ +([0-9]+) +[0-9]+ +[0-9]+ +([0-9]+) +[0-9]+$"
)

# Parse a line from /proc/net/dev.
#
# Example input (includes leading whitespace):
#   eth0: 29819439   19890    0    0    0     0          0         0   364327    6584    0    0    0     0       0          0
PROC_NET_DEV_RE = re.compile(
    r"^ *([^:]+): +([0-9]+) +[0-9]+ +[0-9]+ +[0-9]+ +[0-9]+ +[0-9]+ +[0-9]+ +[0-9]+ +([0-9]+)[0-9 ]+$"
)

# Parse a line from /proc/diskstats
#
# First group is the name. To get partitions rather than disks we require the
# name to end in a number.
#
# Second and third groups are sector reads and writes respectively.
#
# Line format documented here:
# https://www.kernel.org/doc/Documentation/admin-guide/iostats.rst
PROC_DISKSTATS_RE = re.compile(
    r"^ *[0-9]+ +[0-9]+ ([a-z]+[0-9]+) [0-9]+ [0-9]+ ([0-9]+) [0-9]+ [0-9]+ [0-9]+ ([0-9]+) .*"
)


class Sample:
    def __init__(self, name: str, bytecount: int) -> None:
        self.name = name
        self.bytecount = bytecount

    def __repr__(self):
        return f'Sample[name="{self.name}", count={self.bytecount}]'

    def __eq__(self, o):
        return self.bytecount == o.bytecount and self.name == o.name


def parse_netstat_ib_output(netstat_ib_output: str) -> List[Sample]:
    """
    Parse output of "netstat -ib" on macOS.
    """
    samples: List[Sample] = []
    for line in netstat_ib_output.splitlines()[1:]:
        match = NETSTAT_IB_LINE_RE.match(line)
        if not match:
            continue

        incoming_bytes = int(match.group(2))
        outgoing_bytes = int(match.group(3))
        if incoming_bytes == 0 and outgoing_bytes == 0:
            # For our purposes this is just clutter
            continue

        samples.append(Sample(match.group(1) + " incoming", incoming_bytes))
        samples.append(Sample(match.group(1) + " outgoing", outgoing_bytes))

    return samples


def parse_iostat_output(iostat_output: str) -> List[Sample]:
    """
    Parse output of "iostat -dKI -n 99" on macOS.
    """
    lines = iostat_output.splitlines()

    # Example: ["disk0"]
    names = lines[0].split()

    # Example: ["15.95", "2816998", "43889.27"]
    #
    # Numbers are: "KB/t", "xfrs" and "xfrs"
    numbers = lines[2].split()

    samples: List[Sample] = []
    for i, name in enumerate(names):
        mb_string = numbers[3 * i + 2]

        # 1024 * 1024 mirrors what happens in iostat.c:
        # https://opensource.apple.com/source/system_cmds/system_cmds-550.6/iostat.tproj/iostat.c.auto.html
        bytecount = math.trunc(float(mb_string) * 1024 * 1024)

        samples.append(Sample(name, bytecount))

    return samples


def parse_proc_net_dev(proc_net_dev_contents: str) -> List[Sample]:
    """
    Parse /proc/net/dev contents into a list of samples.
    """
    return_me: List[Sample] = []
    for line in proc_net_dev_contents.splitlines():
        match = PROC_NET_DEV_RE.match(line)
        if not match:
            continue

        name = match.group(1)
        incoming = int(match.group(2))
        outgoing = int(match.group(3))
        if incoming == 0 and outgoing == 0:
            continue

        return_me.append(Sample(name + " incoming", incoming))
        return_me.append(Sample(name + " outgoing", outgoing))

    return return_me


def parse_proc_diskstats(proc_diskstats_contents: str) -> List[Sample]:
    """
    Parse /proc/net/dev contents into a list of samples.
    """
    return_me: List[Sample] = []
    for line in proc_diskstats_contents.splitlines():
        match = PROC_DISKSTATS_RE.match(line)
        if not match:
            continue

        name = match.group(1)
        read_sectors = int(match.group(2))
        write_sectors = int(match.group(3))
        if read_sectors == 0 and write_sectors == 0:
            continue

        # Multiply by 512 to get bytes from sectors:
        # https://stackoverflow.com/a/38136179/473672
        return_me.append(Sample(name + " read", read_sectors * 512))
        return_me.append(Sample(name + " write", write_sectors * 512))

    return return_me


class SubsystemStat:
    def __init__(self, throughput: float, high_watermark: float) -> None:
        if throughput > high_watermark:
            raise ValueError(
                f"High watermark {high_watermark} lower than throughput {throughput}"
            )

        self.throughput = throughput
        self.high_watermark = high_watermark


def sample_network_interfaces() -> List[Sample]:
    """
    Query system for network interfaces byte counts
    """

    if os.path.exists("/proc/net/dev"):
        # We're on Linux
        with open("/proc/net/dev", encoding="utf-8") as proc_net_dev:
            return parse_proc_net_dev(proc_net_dev.read())

    # Assuming macOS, add support for more platforms on demand
    netstat_ib_output = px_exec_util.run(["netstat", "-ib"])
    return parse_netstat_ib_output(netstat_ib_output)


def sample_drives() -> List[Sample]:
    """
    Query system for drive statistics
    """

    if os.path.exists("/proc/diskstats"):
        # We're on Linux
        with open("/proc/diskstats", encoding="utf-8") as proc_diskstats:
            return parse_proc_diskstats(proc_diskstats.read())

    # Assuming macOS, add support for more platforms on demand
    iostat_output = px_exec_util.run(["iostat", "-dKI", "-n 99"])
    return parse_iostat_output(iostat_output)


class SystemState:
    def __init__(self) -> None:
        self.timestamp = datetime.datetime.now()

        self.samples: List[Sample] = sample_network_interfaces() + sample_drives()

        by_name: Dict[str, Sample] = {}
        for sample in self.samples:
            by_name[sample.name] = sample
        self.samples_by_name = by_name


class PxIoLoad:
    def __init__(self) -> None:
        self.most_recent_system_state: SystemState = SystemState()
        self.previous_system_state: Optional[SystemState] = None

        # Maps a subsystem name ("eth0 outgoing") to a current bytes-per-second
        # value and a high watermark for the same value.
        self.ios: Dict[str, SubsystemStat] = {}

        # Per interface, keep track of when we first saw it and its byte count
        # at that time.
        self.baseline: Dict[str, Tuple[datetime.datetime, int]] = {}
        self.update_baseline_from_system(self.most_recent_system_state)

    def update_baseline_from_system(self, system_state: SystemState) -> None:
        """
        Add new entries to the baseline, and remove those that aren't part of
        the system state any more.
        """
        unseen = set(self.baseline.keys())
        now = system_state.timestamp
        for sample in system_state.samples:
            if sample.name in unseen:
                unseen.remove(sample.name)

            if sample.name in self.baseline:
                # Already in there, move along!
                continue

            self.baseline[sample.name] = (now, sample.bytecount)

        # Garbage collect removed devices
        for remove_me in unseen:
            del self.baseline[remove_me]

    def update(self) -> None:
        self.previous_system_state = self.most_recent_system_state
        self.most_recent_system_state = SystemState()
        self.update_baseline_from_system(self.most_recent_system_state)

        seconds_since_previous = (
            self.most_recent_system_state.timestamp
            - self.previous_system_state.timestamp
        ).total_seconds()
        assert seconds_since_previous > 0

        # Update self.ios from the system states
        updated_ios: Dict[str, SubsystemStat] = {}
        for sample in self.most_recent_system_state.samples:
            name = sample.name

            baseline_sample = self.baseline[name]
            bytes_since_baseline = sample.bytecount - baseline_sample[1]
            assert bytes_since_baseline >= 0
            seconds_since_baseline = (
                self.most_recent_system_state.timestamp - baseline_sample[0]
            ).total_seconds()
            if seconds_since_baseline == 0:
                # Newly added device, need two samples to make a metric, try
                # again next time
                continue

            previous_sample = self.previous_system_state.samples_by_name.get(name)
            if not previous_sample:
                # Need two samples to make a metric
                continue
            bytes_since_previous = sample.bytecount - previous_sample.bytecount
            assert bytes_since_previous >= 0

            assert bytes_since_baseline >= bytes_since_previous

            bytes_per_second_since_baseline = (
                bytes_since_baseline / seconds_since_baseline
            )
            bytes_per_second_since_previous = (
                bytes_since_previous / seconds_since_previous
            )

            io_entry = self.ios.get(name)
            if not io_entry:
                # New device
                io_entry = SubsystemStat(throughput=0.0, high_watermark=0.0)

            # High watermark throughput should be measured vs the last sample...
            high_watermark = max(
                io_entry.high_watermark, bytes_per_second_since_previous
            )
            assert high_watermark >= bytes_per_second_since_baseline

            # ... but the current B/s should be measured vs the initial sample.
            # This makes the values more stable and easier on the eye.
            updated_ios[name] = SubsystemStat(
                throughput=bytes_per_second_since_baseline,
                high_watermark=high_watermark,
            )

        self.ios = updated_ios

    def get_load_string(self) -> str:
        """
        Example return value: "[123B/s / 878B/s] eth0 outgoing"
        """

        # NOTE: To compute this value, we need a collection of data points, with
        # each data point containing:
        # * A name ("eth0 outgoing")
        # * A current bytes-per-second value (42.1)
        # * The high watermark value
        #
        # This data is available in self.ios.
        #
        # Then, we need to sort these by current-bytes-per-second, and render the top one.

        # Values per entry: name, current value, high watermark
        collected_ios: List[Tuple[str, float, float]] = []
        for name, loads in self.ios.items():
            collected_ios.append((name, loads.throughput, loads.high_watermark))

        if not collected_ios:
            # No load collected
            return ""

        # Highest throughput first
        collected_ios.sort(key=lambda collectee: collectee[1], reverse=True)

        bottleneck = collected_ios[0]
        current_throughput, max_throughput = px_units.bytes_to_strings(
            math.trunc(bottleneck[1]), math.trunc(bottleneck[2])
        )

        return "[{} / {}] {}".format(
            px_terminal.bold(current_throughput + "/s"),
            max_throughput + "/s",
            px_terminal.bold(bottleneck[0]),
        )


_ioload = PxIoLoad()


def update():
    _ioload.update()


def get_load_string():
    return _ioload.get_load_string()



import errno
import platform

from . import px_exec_util

import sys

from typing import Optional
from typing import Tuple
from typing import List


def get_core_count() -> Tuple[int, int]:
    """
    Count the number of cores in the system.

    Returns a tuple (physical, logical) with counts of physical and logical
    cores.
    """
    return_me = get_core_count_from_proc_cpuinfo()
    if return_me is not None:
        return return_me

    return_me = get_core_count_from_sysctl()
    if return_me is not None:
        return return_me

    uname = str(platform.uname())
    platform_s = uname + " Python " + sys.version

    raise IOError("Unable to get cores info " + platform_s)


def get_core_count_from_proc_cpuinfo(
    proc_cpuinfo: str = "/proc/cpuinfo",
) -> Optional[Tuple[int, int]]:
    """
    Count the number of cores in /proc/cpuinfo.

    Returns a tuple (physical, logical) with counts of physical and logical
    cores.
    """
    # Note the ending spaces, they must be there for number extraction to work!
    PROCESSOR_NO_PREFIX = "processor\t: "
    CORE_ID_PREFIX = "core id\t\t: "

    core_ids = set()
    max_processor_no = 0
    try:
        with open(proc_cpuinfo, encoding="utf-8") as f:
            for line in f:
                if line.startswith(PROCESSOR_NO_PREFIX):
                    processor_no = int(line[len(PROCESSOR_NO_PREFIX) :])
                    max_processor_no = max(processor_no, max_processor_no)
                elif line.startswith(CORE_ID_PREFIX):
                    core_id = int(line[len(CORE_ID_PREFIX) :])
                    core_ids.add(core_id)
    except (IOError, OSError) as e:
        if e.errno == errno.ENOENT:
            # /proc/cpuinfo not found, we're probably not on Linux
            return None

        raise

    physical = len(core_ids)
    logical = max_processor_no + 1
    if physical == 0:
        # I get this on my cell phone
        physical = logical
    return (physical, logical)


def get_core_count_from_sysctl() -> Optional[Tuple[int, int]]:
    try:
        stdout = px_exec_util.run(["sysctl", "hw"])
    except (IOError, OSError) as e:
        if e.errno == errno.ENOENT:
            # sysctl not found, we're probably not on OSX
            return None

        raise

    sysctl_lines = stdout.splitlines()

    physical, logical = parse_sysctl_output(sysctl_lines)
    if physical is None or logical is None:
        # On Linux, sysctl exists but it doesn't contain the values we want
        return None

    return (physical, logical)


def parse_sysctl_output(
    sysctl_lines: List[str],
) -> Tuple[Optional[int], Optional[int]]:
    # Note the ending spaces, they must be there for number extraction to work!
    PHYSICAL_PREFIX = "hw.physicalcpu: "
    LOGICAL_PREFIX = "hw.logicalcpu: "

    physical = None
    logical = None
    for line in sysctl_lines:
        if line.startswith(PHYSICAL_PREFIX):
            physical = int(line[len(PHYSICAL_PREFIX) :])
        elif line.startswith(LOGICAL_PREFIX):
            logical = int(line[len(LOGICAL_PREFIX) :])

    return (physical, logical)



import socket

from . import px_exec_util

from typing import Set
from typing import List
from typing import Tuple
from typing import Optional


class PxFile:
    def __init__(self, pid: int, filetype: str) -> None:
        self.fd: Optional[int] = None
        self.pid = pid
        self.type = filetype
        self.name: Optional[str] = None
        self.inode: Optional[str] = None
        self.device: Optional[str] = None
        self.access: Optional[str] = None

        # Example values: "cwd", "txt" and probably others as well
        self.fdtype: Optional[str] = None

    def __repr__(self):
        # The point of implementing this method is to make the py.test output
        # more readable.
        return str(self.pid) + ":" + str(self)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((self.name, self.fd, self.fdtype, self.pid))

    def __str__(self):
        if self.type == "REG":
            return self.name

        name = self.name
        listen_suffix = ""
        if self.type in ["IPv4", "IPv6"]:
            _, remote_endpoint = self.get_endpoints()
            if not remote_endpoint:
                listen_suffix = " (LISTEN)"

            name = self._resolve_name()

        # Decorate non-regular files with their type
        if name:
            return "[" + self.type + "] " + name + listen_suffix
        return "[" + self.type + "] " + listen_suffix

    def _resolve_name(self):
        local, remote = self.get_endpoints()
        if not local:
            return self.name

        local = resolve_endpoint(local)
        if not remote:
            return local

        return local + "->" + resolve_endpoint(remote)

    def device_number(self):
        if self.device is None:
            return None

        number = int(self.device, 16)
        if number == 0:
            # 0x000lotsofmore000 is what we get on lsof 4.86 and Linux 4.2.0
            # when lsof doesn't have root privileges
            return None

        return number

    def fifo_id(self):
        if self.inode is not None:
            # On Linux, pipes are presented by lsof as FIFOs. They have unique-
            # per-pipe inodes, so we use them as IDs.
            return self.inode

        if self.type == "FIFO" and self.name == "pipe":
            # This is just a label that can be shared by several pipes on Linux,
            # we can't use it to identify a pipe.
            return None

        if self.type == "PIPE" and self.name == "(none)":
            # This is just a label that can be shared by several pipes on OS X,
            # we can't use it to identify a pipe.
            return None

        # On OS X, pipes are presented as PIPEs and lack inodes, but they
        # compensate by having unique names.
        return self.name

    def get_endpoints(self) -> Tuple[Optional[str], Optional[str]]:
        """
        Returns a (local,remote) tuple. They represent the local and the remote
        endpoints of a network connection.

        This method will never return None, but both local and remote can be
        None in case this isn't a network connection for example.
        """
        if not self.name:
            return (None, None)

        if self.type not in ["IPv4", "IPv6"]:
            return (None, None)

        local = None
        remote = None

        split_name = self.name.split("->")
        local = split_name[0]

        # Turn "localhost:ipp (LISTEN)" into "ipp" and nothing else
        local = local.split(" ")[0]
        if "*" in local:
            # We can't match against this endpoint
            local = None

        if len(split_name) == 2:
            remote = split_name[1]

        return (local, remote)


class PxFileBuilder:
    def __init__(self):
        self.fd: Optional[int] = None
        self.pid: Optional[int] = None
        self.name: Optional[str] = None
        self.type: Optional[str] = None
        self.inode: Optional[str] = None
        self.device: Optional[str] = None
        self.access: Optional[str] = None

        # Example values: "cwd", "txt" and probably others as well
        self.fdtype: Optional[str] = None

    def build(self) -> PxFile:
        assert self.pid is not None
        assert self.type
        pxFile = PxFile(self.pid, self.type)
        pxFile.name = self.name
        pxFile.fd = self.fd
        pxFile.inode = self.inode
        pxFile.device = self.device
        pxFile.access = self.access
        pxFile.fdtype = self.fdtype

        return pxFile

    def __repr__(self):
        return f"PxFileBuilder(pid={self.pid}, name={self.name}, type={self.type})"


def resolve_endpoint(endpoint: str) -> str:
    """
    Resolves "127.0.0.1:portnumber" into "localhost:portnumber".
    """
    # Find the rightmost :, necessary for IPv6 addresses
    splitindex = endpoint.rfind(":")
    if splitindex == -1:
        return endpoint

    address = endpoint[0:splitindex]
    if address[0] == "[" and address[-1] == "]":
        # This is how lsof presents IPv6 addresses
        address = address[1:-1]

    port = endpoint[splitindex + 1 :]
    host = None

    try:
        host = socket.gethostbyaddr(address)[0]
    except Exception:
        # Lookup failed for whatever reason, give up
        #
        # Catching "Exception" because I am (on 2022may27) unable to figure out
        # from these docs what exceptions can be thrown by that method, if any:
        # https://docs.python.org/3.10/library/socket.html#socket.gethostbyaddr
        return endpoint

    if host == "localhost.localdomain":
        # "localdomain" is just a long word that doesn't add any information
        host = "localhost"

    return host + ":" + port


def call_lsof():
    """
    Call lsof and return the result as one big string
    """
    # See OUTPUT FOR OTHER PROGRAMS: http://linux.die.net/man/8/lsof
    # Output lines can be in one of two formats:
    # 1. "pPID@" (with @ meaning NUL)
    # 2. "fFD@aACCESSMODE@tTYPE@nNAME@"
    return px_exec_util.run(["lsof", "-n", "-F", "fnaptd0i"])


def lsof_to_files(lsof: str) -> List[PxFile]:
    """
    Convert lsof output into a files array.
    """

    pid = None
    file_builder: Optional[PxFileBuilder] = None
    files: List[PxFile] = []
    for shard in lsof.split("\0"):
        if shard[0] == "\n":
            # Some shards start with newlines. Looks pretty when viewing the
            # lsof output in moar, but makes the parsing code have to deal with
            # it.
            shard = shard[1:]

        if not shard:
            # The output ends with a single newline, which we just stripped away
            break

        infotype = shard[0]
        value = shard[1:]

        if infotype == "p":
            pid = int(value)
        elif infotype == "f":
            if file_builder:
                files.append(file_builder.build())
            else:
                file_builder = PxFileBuilder()

            # Reset the file builder object. This operation is on a hot path
            # and doing without an extra object allocation here actually helps.
            file_builder.__init__()  # type: ignore

            if value.isdigit():
                file_builder.fd = int(value)
            else:
                # Words like "cwd", "txt" and probably others as well
                file_builder.fdtype = value
            assert pid is not None
            file_builder.pid = pid
            file_builder.type = "??"
        elif infotype == "a":
            assert file_builder is not None
            access = {" ": None, "r": "r", "w": "w", "u": "rw"}[value]
            file_builder.access = access
        elif infotype == "t":
            assert file_builder is not None
            file_builder.type = value
        elif infotype == "d":
            assert file_builder is not None
            file_builder.device = value
        elif infotype == "n":
            assert file_builder is not None
            file_builder.name = value
        elif infotype == "i":
            assert file_builder is not None
            file_builder.inode = value

        else:
            raise Exception(f"Unhandled type <{infotype}> for shard <{shard}>")

    if file_builder:
        # Don't forget the last file
        files.append(file_builder.build())

    return files


def get_all() -> Set[PxFile]:
    """
    Get all files.

    If a file_types array is specified, only files of the listed types will be
    returned.
    """
    return set(lsof_to_files(call_lsof()))



import logging
import datetime

import re

from . import px_exec_util

from typing import Optional, Set

LOG = logging.getLogger(__name__)

TIMEZONE = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo

# last regexp parts
LAST_USERNAME = "([^ ]+)"
LAST_DEVICE = "([^ ]+)"
LAST_ADDRESS = "([^ ]+)?"
LAST_PID = r"( \[[0-9]+\])?"
LAST_FROM = "(... ... .. ..:..)"
LAST_DASH = " [- ] "
LAST_TO = "[^(]*"
LAST_DURATION = "([0-9+:]+)"

LAST_RE = re.compile(
    LAST_USERNAME
    + " +"
    + LAST_DEVICE
    + " +"
    + LAST_ADDRESS
    + LAST_PID
    + " +"
    + LAST_FROM
    + LAST_DASH
    + LAST_TO
    + r" *(\("
    + LAST_DURATION
    + r"\))?"
)

TIMEDELTA_RE = re.compile(r"(([0-9]+)\+)?([0-9][0-9]):([0-9][0-9])")

# Month names in English locale
MONTHS = {
    "Jan": 1,
    "Feb": 2,
    "Mar": 3,
    "Apr": 4,
    "May": 5,
    "Jun": 6,
    "Jul": 7,
    "Aug": 8,
    "Sep": 9,
    "Oct": 10,
    "Nov": 11,
    "Dec": 12,
}


def get_users_at(
    timestamp: datetime.datetime,
    last_output: Optional[str] = None,
    now: Optional[datetime.datetime] = None,
) -> Set[str]:
    """
    Return a set of strings corresponding to which users were logged in from
    which addresses at a given timestamp.

    Optional argument last_output is the output of "last". Will be filled in by
    actually executing "last" if not provided.

    Optional argument now is the current timestamp for parsing last_output. Will
    be taken from the system clock if not provided.
    """

    if now is None:
        now = datetime.datetime.now(datetime.timezone.utc).astimezone()

    if last_output is None:
        last_output = call_last()

    users = set()
    for line in last_output.splitlines():
        if not line:
            continue
        if line.startswith("wtmp begins"):
            # This is trailing noise printed by last
            continue
        if line.startswith("reboot "):
            continue
        if line.startswith("shutdown "):
            continue

        match = LAST_RE.match(line)
        if not match:
            LOG.error("Unmatched last line: <%s>", line)
            continue

        username = match.group(1)
        address = match.group(3)
        from_s = match.group(5)
        duration_s = match.group(7)

        if address:
            username += " from " + address

        try:
            from_timestamp = _to_timestamp(from_s, now)
            if timestamp < from_timestamp:
                continue
        except Exception:
            LOG.error("Problematic1 last line: <%s>", line)
            continue

        if duration_s is None:
            # Still logged in
            users.add(username)
            continue

        try:
            duration_delta = _to_timedelta(duration_s)
            to_timestamp = from_timestamp + duration_delta
            if timestamp > to_timestamp:
                continue
        except Exception:
            LOG.error("Problematic2 last line: <%s>", line)

        users.add(username)

    return users


def call_last():
    """
    Call last and return the result as one big string
    """
    return px_exec_util.run(["last"])


def _to_timestamp(string, now):
    """
    Parse a timestamp like "Mon Feb  7 13:19".

    Names of months and days must be in English, make sure you call "last"
    without the LANG= environment variable set.

    Now is the current timestamp, and the returned timestamp is supposed to be
    before now.
    """

    # FIXME: Use day of week as a checksum before returning from this function?

    split = string.split()
    month = MONTHS[split[1]]
    day = int(split[2])

    hour, minute = split[3].split(":")
    hour = int(hour)
    minute = int(minute)

    try:
        timestamp = datetime.datetime(
            now.year, month, day, hour, minute, tzinfo=TIMEZONE
        )
        if timestamp <= now:
            return timestamp
    except ValueError:
        if month == 2 and day == 29:
            # This happens at leap years when we get the year wrong
            pass
        else:
            raise

    return datetime.datetime(now.year - 1, month, day, hour, minute, tzinfo=TIMEZONE)


def _to_timedelta(string: str) -> datetime.timedelta:
    match = TIMEDELTA_RE.match(string)
    assert match is not None

    matched_days = match.group(2)
    if matched_days is None:
        days = 0
    else:
        days = int(matched_days)

    hours = int(match.group(3))

    minutes = int(match.group(4))

    return datetime.timedelta(days, hours=hours, minutes=minutes)



import logging
import datetime
import operator

import os
import re
import pwd
import errno
import subprocess

from . import px_commandline
from . import px_exec_util


from typing import Dict
from typing import Optional
from typing import List
from typing import Iterable


LOG = logging.getLogger(__name__)


# Match + group: " 7708 1 Mon Mar  7 09:33:11 2016  netbios 0.1 0:00.08  0.0 /usr/sbin/netbiosd hj"
PS_LINE = re.compile(
    " *([0-9]+) +([0-9]+) +([0-9]+) +([A-Za-z0-9: ]+) +([^ ]+) +([0-9.]+) +([-0-9.:]+) +([0-9.]+) +(.*)"
)

# Match + group: "1:02.03"
CPUTIME_OSX = re.compile(r"^([0-9]+):([0-9][0-9]\.[0-9]+)$")

# Match + group: "01:23:45"
CPUTIME_LINUX = re.compile("^([0-9][0-9]):([0-9][0-9]):([0-9][0-9])$")

# Match + group: "123-01:23:45"
CPUTIME_LINUX_DAYS = re.compile("^([0-9]+)-([0-9][0-9]):([0-9][0-9]):([0-9][0-9])$")


TIMEZONE = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo


uid_to_username_cache: Dict[int, str] = {}
get_command_cache: Dict[str, str] = {}


def _parse_time(time_s: str) -> datetime.datetime:
    """
    Parse a local date from ps into a datetime.datetime object.

    Example inputs:
      "Wed Dec 16 12:41:43 2020"
      "Sat Jan  9 14:20:34 2021"
    """

    zero_based_month = [
        "Jan",
        "Feb",
        "Mar",
        "Apr",
        "May",
        "Jun",
        "Jul",
        "Aug",
        "Sep",
        "Oct",
        "Nov",
        "Dec",
    ].index(time_s[4:7])

    day_of_month = int(time_s[8:10])
    hour = int(time_s[11:13])
    minute = int(time_s[14:16])
    second = int(time_s[17:19])
    year = int(time_s[20:24])

    return datetime.datetime(
        year, zero_based_month + 1, day_of_month, hour, minute, second, tzinfo=TIMEZONE
    )


class PxProcess:
    def __init__(
        self,
        cmdline: str,
        pid: int,
        rss_kb: int,
        start_time_string: str,
        username: str,
        now: datetime.datetime,
        ppid: Optional[int],
        memory_percent: Optional[float] = None,
        cpu_percent: Optional[float] = None,
        cpu_time: Optional[float] = None,
        aggregated_cpu_time: float = 0.0,
    ) -> None:
        self.pid: int = pid
        self.ppid: Optional[int] = ppid
        self.rss_kb: int = rss_kb

        self.cmdline: str = cmdline
        self.command: str = self._get_command()
        self.lowercase_command: str = self.command.lower()

        self.start_time = _parse_time(start_time_string.strip())
        self.age_seconds: float = (now - self.start_time).total_seconds()
        if self.age_seconds < 0:
            self.age_seconds = 0

        self.age_s: str = seconds_to_str(self.age_seconds)

        self.username: str = username

        self.memory_percent = memory_percent
        self.memory_percent_s: str = "--"
        if memory_percent is not None:
            self.memory_percent_s = f"{memory_percent:.0f}%"

        self.cpu_percent = cpu_percent
        self.cpu_percent_s: str = "--"
        if cpu_percent is not None:
            self.cpu_percent_s = f"{cpu_percent:.0f}%"

        self.set_cpu_time_seconds(cpu_time)
        self.set_aggregated_cpu_time_seconds(aggregated_cpu_time)

        self.children: List[PxProcess] = []
        self.parent: Optional[PxProcess] = None

        # How many levels down the tree this process is. Kernel is level 0, init
        # level 1 and everything else 2 and up.
        self.level = 0

    def __repr__(self):
        # I guess this is really what __str__ should be doing, but the point of
        # implementing this method is to make the py.test output more readable,
        # and py.test calls repr() and not str().
        return str(self.pid) + ":" + self.command

    def __str__(self):
        return self.command + "(" + str(self.pid) + ")"

    def __eq__(self, other):
        if other is None:
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return self.pid

    def set_cpu_time_seconds(self, seconds: Optional[float]) -> None:
        self.cpu_time_s: str = "--"
        self.cpu_time_seconds = None
        if seconds is not None:
            self.cpu_time_s = seconds_to_str(seconds)
            self.cpu_time_seconds = seconds

    def set_aggregated_cpu_time_seconds(self, seconds: float) -> None:
        self.aggregated_cpu_time_s = seconds_to_str(seconds)
        self.aggregated_cpu_time_seconds = seconds

    def match(self, string, require_exact_user=True):
        """
        Returns True if this process matches the string.

        See px_process_test.test_match() for the exact definition of how the
        matching is done.
        """
        if not string:
            return True

        if self.username == string:
            return True

        if not require_exact_user:
            if self.username.startswith(string):
                return True

        if string in self.cmdline:
            return True

        if string in self.cmdline.lower():
            return True

        if str(self.pid).startswith(string):
            return True

        return False

    def get_command_line_array(self):
        return px_commandline.to_array(self.cmdline)

    def _get_command(self):
        """Return just the command without any arguments or path"""
        if self.cmdline in get_command_cache:
            return get_command_cache[self.cmdline]

        command = px_commandline.get_command(self.cmdline)
        get_command_cache[self.cmdline] = command

        return command

    def get_sudo_user(self):
        """Retrieves the $SUDO_USER value for this process, or None if not set"""
        stdout = px_exec_util.run(["ps", "e", str(self.pid)])
        match = re.match(".* SUDO_USER=([^ ]+)", stdout, re.DOTALL)
        if not match:
            return None

        return match.group(1)

    def is_alive(self):
        try:
            # Signal 0 has no effect
            os.kill(self.pid, 0)
        except OSError as e:
            if e.errno == errno.ESRCH:
                # No such process
                return False

            # Process found but something else went wrong
            return True

        # No problem, process was there
        return True


class PxProcessBuilder:
    def __init__(self):
        self.cmdline: Optional[str] = None
        self.pid: Optional[int] = None
        self.ppid: Optional[int] = None
        self.rss_kb: Optional[int] = None
        self.start_time_string: Optional[str] = None
        self.username: Optional[str] = None
        self.cpu_percent: Optional[float] = None
        self.cpu_time: Optional[float] = None
        self.memory_percent: Optional[float] = None

    def __repr__(self):
        return (
            "start_time_string=%r pid=%r ppid=%r user=%r cpu%%=%r cputime=%r mem%%=%r cmd=<%r>"
            % (
                self.start_time_string,
                self.pid,
                self.ppid,
                self.username,
                self.cpu_percent,
                self.cpu_time,
                self.memory_percent,
                self.cmdline,
            )
        )

    def build(self, now: datetime.datetime) -> PxProcess:
        assert self.cmdline
        assert self.pid is not None
        assert self.rss_kb is not None
        assert self.start_time_string
        assert self.username
        return PxProcess(
            cmdline=self.cmdline,
            pid=self.pid,
            ppid=self.ppid,
            rss_kb=self.rss_kb,
            start_time_string=self.start_time_string,
            username=self.username,
            now=now,
            memory_percent=self.memory_percent,
            cpu_percent=self.cpu_percent,
            cpu_time=self.cpu_time,
        )


def parse_time(timestring: str) -> float:
    """Convert a CPU time string returned by ps to a number of seconds"""

    match = CPUTIME_OSX.match(timestring)
    if match:
        minutes = int(match.group(1))
        seconds = float(match.group(2))
        return 60 * minutes + seconds

    match = CPUTIME_LINUX.match(timestring)
    if match:
        hours = int(match.group(1))
        minutes = int(match.group(2))
        seconds = int(match.group(3))
        return 3600 * hours + 60 * minutes + seconds

    match = CPUTIME_LINUX_DAYS.match(timestring)
    if match:
        days = int(match.group(1))
        hours = int(match.group(2))
        minutes = int(match.group(3))
        seconds = int(match.group(4))
        return 86400 * days + 3600 * hours + 60 * minutes + seconds

    raise ValueError("Unparsable timestamp: <" + timestring + ">")


def uid_to_username(uid: int) -> str:
    if uid in uid_to_username_cache:
        return uid_to_username_cache[uid]

    # Populate cache
    try:
        uid_to_username_cache[uid] = str(pwd.getpwuid(uid).pw_name)
    except KeyError:
        uid_to_username_cache[uid] = str(uid)

    return uid_to_username_cache[uid]


def ps_line_to_process(ps_line: str, now: datetime.datetime) -> PxProcess:
    match = PS_LINE.match(ps_line)
    if not match:
        raise Exception(f"Failed to match ps line <{ps_line:!r}>")

    process_builder = PxProcessBuilder()
    process_builder.pid = int(match.group(1))
    process_builder.ppid = int(match.group(2))
    process_builder.rss_kb = int(match.group(3))
    process_builder.start_time_string = match.group(4)
    process_builder.username = uid_to_username(int(match.group(5)))
    process_builder.cpu_percent = float(match.group(6))
    process_builder.cpu_time = parse_time(match.group(7))
    process_builder.memory_percent = float(match.group(8))
    process_builder.cmdline = match.group(9)

    return process_builder.build(now)


def create_kernel_process(now: datetime.datetime) -> PxProcess:
    """
    Fake a process 0, this one isn't returned by ps. More info about PID 0:
    https://en.wikipedia.org/wiki/Process_identifier
    """
    process_builder = PxProcessBuilder()
    process_builder.pid = 0
    process_builder.ppid = None

    # FIXME: This should be the system boot timestamp, not the epoch
    process_builder.start_time_string = datetime.datetime.fromtimestamp(
        0, datetime.timezone.utc
    ).strftime("%c")

    process_builder.rss_kb = 0
    process_builder.username = "root"
    process_builder.cpu_time = None
    process_builder.memory_percent = None
    process_builder.cmdline = "kernel PID 0"
    process = process_builder.build(now)

    return process


def resolve_links(processes: Dict[int, PxProcess], now: datetime.datetime) -> None:
    """
    On entry, this function assumes that all processes have a "ppid" field
    containing the PID of their parent process.

    When done, all processes will have a "parent" field with a reference to the
    process' parent process object.

    Also, all processes will have a (possibly empty) "children" field containing
    a set of references to child processes.
    """
    if 0 not in processes:
        kernel_process = create_kernel_process(now)
        processes[0] = kernel_process

    for process in processes.values():
        if process.pid == 0:
            process.parent = None
        elif process.ppid is None:
            process.parent = None
        else:
            process.parent = processes.get(process.ppid)

        if process.parent is not None:
            process.parent.children.append(process)


def remove_process_and_descendants(processes: Dict[int, PxProcess], pid: int) -> None:
    process = processes[pid]
    if process.parent is not None:
        process.parent.children.remove(process)
    toexclude = [process]
    while toexclude:
        process = toexclude.pop()
        del processes[process.pid]
        for child in process.children:
            toexclude.append(child)


def get_all() -> List[PxProcess]:
    processes = {}

    # NOTE: Both the full path to ps and "close_fds = False" are important
    # because they enable the use of _posix_spawn()...
    #
    # https://github.com/python/cpython/blob/998ae1fa3fb05a790071217cf8f6ae3a928da13f/Lib/subprocess.py#L1715
    #
    # ... and avoids prematurely waiting for ps to produce 50000 bytes:
    #
    # https://github.com/python/cpython/blob/998ae1fa3fb05a790071217cf8f6ae3a928da13f/Lib/subprocess.py#L1796
    #
    # If you want to change this, try benchmark_proc_get_all.py and make sure
    # you don't regress.
    close_fds = False
    command = [
        "/bin/ps",
        "-ax",
        "-o",
        "pid=,ppid=,rss=,lstart=,uid=,pcpu=,time=,%mem=,command=",
    ]

    with open(os.devnull, "w", encoding="utf-8") as DEVNULL:
        with subprocess.Popen(
            command,
            stdin=DEVNULL,
            stdout=subprocess.PIPE,
            stderr=DEVNULL,
            close_fds=close_fds,
            env=px_exec_util.ENV,
        ) as ps:
            stdout = ps.stdout
            assert stdout
            now = datetime.datetime.now().replace(tzinfo=TIMEZONE)
            for ps_line in stdout:
                process = ps_line_to_process(ps_line.decode("utf-8"), now)
                processes[process.pid] = process

            if ps.wait() != 0:
                raise OSError(f"Exit code {ps.returncode} from {command}")

    resolve_links(processes, now)
    remove_process_and_descendants(processes, os.getpid())

    return list(processes.values())


def order_best_last(processes: Iterable[PxProcess]) -> List[PxProcess]:
    """
    Returns process list ordered with the most interesting one last.

    The interestingness of a process is determined by three factors:
    * How recently it was started. More recently is more interesting.
    * How much memory it is using. Higher memory usage is more interesting.
    * How much CPU it is using right now. Higher CPU usage is more interesting.
    * How much CPU time it has used. Higher CPU time is more interesting.

    To sort by all three of these factors, we first make three lists. The score
    of each process is determined by its best position in any of the lists.
    """

    # Earlier in these lists = more interesting
    by_age = sorted(
        filter(lambda p: p.age_seconds is not None, processes),
        key=operator.attrgetter("age_seconds"),
    )
    by_memory = sorted(
        filter(lambda p: p.memory_percent is not None, processes),
        key=operator.attrgetter("memory_percent"),
        reverse=True,
    )
    by_cpu = sorted(
        filter(lambda p: p.cpu_percent is not None, processes),
        key=operator.attrgetter("cpu_percent"),
        reverse=True,
    )
    by_cpu_time = sorted(
        filter(lambda p: p.cpu_time_seconds is not None, processes),
        key=operator.attrgetter("cpu_time_seconds"),
        reverse=True,
    )

    age_to_position: Dict[int, int] = {}
    for position, process in enumerate(by_age):
        age_to_position[process.pid] = position

    memory_to_position: Dict[float, int] = {}
    for position, process in enumerate(by_memory):
        memory_to_position[process.pid] = position

    cpu_to_position: Dict[float, int] = {}
    for position, process in enumerate(by_cpu):
        cpu_to_position[process.pid] = position

    cpu_time_to_position: Dict[float, int] = {}
    for position, process in enumerate(by_cpu_time):
        cpu_time_to_position[process.pid] = position

    # Presumably higher than the number of processes on any system
    NOT_FOUND = 99999

    def score(process: PxProcess) -> int:
        """Higher score = more interesting"""
        best_position = min(
            age_to_position.get(process.pid, NOT_FOUND),
            memory_to_position.get(process.pid, NOT_FOUND),
            cpu_to_position.get(process.pid, NOT_FOUND),
            cpu_time_to_position.get(process.pid, NOT_FOUND),
        )
        return -best_position

    return sorted(processes, key=score)


def order_best_first(processes: Iterable[PxProcess]) -> List[PxProcess]:
    """Returns process list ordered with the most interesting one first"""
    return order_best_last(processes)[::-1]


def seconds_to_str(seconds: float) -> str:
    if seconds < 60:
        seconds_s = str(seconds)
        decimal_index = seconds_s.rfind(".")
        if decimal_index > -1:
            # Chop to at most two decimals
            seconds_s = seconds_s[0 : decimal_index + 3]
        return seconds_s + "s"

    if seconds < 3600:
        minutes = int(seconds / 60)
        remaining_seconds = int(seconds - minutes * 60)
        return f"{minutes}m{remaining_seconds:02d}s"

    if seconds < 86400:
        hours = int(seconds / 3600)
        minutes = int((seconds - 3600 * hours) / 60)
        return f"{hours}h{minutes:02d}m"

    days = int(seconds / 86400)
    hours = int((seconds - 86400 * days) / 3600)
    return f"{days}d{hours:02d}h"



import sys
import shutil

import os


def install(src, dest):
    """
    Copy src (file) into dest (file) and make dest executable.

    On trouble, prints message and exits with an error code.
    """
    try:
        _install(src, dest)
    except Exception as e:
        sys.stderr.write(f"Installing {dest} failed, please retry with sudo\n")
        sys.stderr.write(f"Error was: {str(e)}\n")
        sys.exit(1)
    print(f"Created: {dest}")


def _install(src, dest):
    """
    Copy src (file) into dest (file) and make dest executable.

    Throws exception on trouble.
    """
    if os.path.realpath(src) == os.path.realpath(dest):
        # Copying a file onto itself is a no-op, never mind
        return

    if not os.path.isfile(src):
        raise OSError(f"Source is not a file: {src}")

    parent = os.path.dirname(dest)
    if not os.path.isdir(parent):
        raise OSError(f"Destination parent is not a directory: {parent}")

    if os.path.isdir(dest):
        raise OSError(f"Destination is a directory, won't replace that: {dest}")

    # Make sure nothing's in the way
    try:
        os.remove(dest)
    except OSError:
        pass
    if os.path.exists(dest):
        raise OSError(f"Can't remove existing entry: {dest}")

    shutil.copyfile(src, dest)
    os.chmod(dest, 0o755)



"""Extract information from command lines"""

import re
import os.path
import logging

from typing import List, Optional, Callable

LOG = logging.getLogger(__name__)


# Match "[kworker/0:0H]", no grouping
LINUX_KERNEL_PROC = re.compile("^\\[[^/ ]+/?[^/ ]+\\]$")

# Match "(python2.7)", no grouping
OSX_PARENTHESIZED_PROC = re.compile("^\\([^()]+\\)$")

# Name of the Perl interpreter
PERL_BIN = re.compile("^perl[.0-9]*$")


def get_trailing_absolute_path(so_far: str) -> Optional[str]:
    """
    Extract a potential file path from the end of a string.

    The coalescing logic will then base decisions on whether this file path
    exists or not.
    """
    start_index = -1
    if so_far.startswith("/"):
        start_index = 0
    elif so_far.startswith("-"):
        equals_slash_index = so_far.find("=/")
        if equals_slash_index == -1:
            # No =/ in the string, so we're not looking at -Djava.io.tmpdir=/tmp
            return None

        # Start at the slash after the equals sign
        start_index = equals_slash_index + 1

    if start_index == -1:
        # Start not found, this is not a path
        return None

    colon_slash_index = so_far.rfind(":/", start_index)
    if colon_slash_index == -1:
        # Not a : separated path
        return so_far[start_index:]

    return so_far[colon_slash_index + 1 :]


def should_coalesce(
    parts: List[str], exists: Callable[[str], bool] = os.path.exists
) -> Optional[bool]:
    """
    Two or more (previously) space separated command line parts should be
    coalesced if combining them with a space in between creates an existing file
    path, or a : separated series of file paths.

    Return values:
    * True: Coalesce, done
    * False: Do not coalesce, done
    * None: Undecided, add another part and try again
    """

    if parts[-1].startswith("-") or parts[-1].startswith("/"):
        # Last part starts a command line option or a new absolute path, don't
        # coalesce.
        return False

    coalesced = " ".join(parts)
    candidate = get_trailing_absolute_path(coalesced)
    if not candidate:
        # This is not a candidate for coalescing
        return False

    if exists(candidate):
        # Found it, done!
        return True

    if exists(os.path.dirname(candidate)):
        # Found the parent directory, we're on the right track, keep looking!
        return None

    # Candidate does not exists, and neither does its parent directory, this is
    # not it.
    return False


def coalesce_count(
    parts: List[str], exists: Callable[[str], bool] = os.path.exists
) -> int:
    """How many parts should be coalesced?"""

    for coalesce_count in range(2, len(parts) + 1):
        should_coalesce_ = should_coalesce(parts[0:coalesce_count], exists)

        if should_coalesce_ is None:
            # Undecided, keep looking
            continue

        if should_coalesce_ is False:
            return 1

        # should_coalesce_ is True
        return coalesce_count

    # Undecided until the end, this means no coalescing should be done
    return 1


def to_array(
    commandline: str, exists: Callable[[str], bool] = os.path.exists
) -> List[str]:
    """Splits a command line string into components"""
    base_split = commandline.split(" ")
    if len(base_split) == 1:
        return base_split

    # Try to reverse engineer which spaces should not be split on
    merged_split: List[str] = []
    i = 0
    while i < len(base_split):
        coalesce_count_ = coalesce_count(base_split[i:], exists)
        merged_split.append(" ".join(base_split[i : i + coalesce_count_]))
        i += coalesce_count_

    return merged_split


def is_human_friendly(command: str) -> bool:
    """
    AKA "Does this command contain any capital letters?"
    """
    for char in command:
        if char.isupper():
            return True
    return False


def get_app_name_prefix(commandline: str) -> str:
    """
    On macOS, get which app this command is part of.
    """
    command_with_path = to_array(commandline)[0]
    command = os.path.basename(command_with_path)
    for part in command_with_path.split("/"):
        if "." not in part:
            continue

        name, suffix = part.rsplit(".", 1)
        if suffix not in ["app", "framework"]:
            continue

        if name == command:
            continue

        return name + "/"

    return ""


def try_clarify_electron(commandline: str) -> Optional[str]:
    for path_component in to_array(commandline)[0].split("/"):
        if path_component.endswith(".app"):
            return path_component[:-4]

    return None


def faillog(commandline: str, parse_result: Optional[str]) -> str:
    """
    If successful, just return the result. If unsuccessful log the problem and
    return the VM name.
    """
    if parse_result:
        return parse_result

    LOG.debug("Parsing failed, using fallback: <%s>", commandline)

    vm = os.path.basename(to_array(commandline)[0])
    return vm


def get_command(commandline: str) -> str:
    """
    Extracts the command from the command line.

    This function most often returns the first component of the command line
    with the path stripped away.

    For some language runtimes, this function may return the name of the program
    that the runtime is executing.
    """
    if LINUX_KERNEL_PROC.match(commandline):
        return commandline

    if OSX_PARENTHESIZED_PROC.match(commandline):
        return commandline

    command = os.path.basename(to_array(commandline)[0])

    if command.startswith("python") or command == "Python":
        return faillog(commandline, get_python_command(commandline))

    if command.startswith("guile"):
        return faillog(commandline, get_guile_command(commandline))

    if command == "Electron":
        clarified = try_clarify_electron(commandline)
        if clarified:
            return clarified

    if command == "java":
        return faillog(commandline, get_java_command(commandline))

    if command == "dart":
        return faillog(commandline, get_dart_command(commandline))

    if command == "ruby":
        # Switches list inspired by ruby 2.3.7p456 --help output
        return faillog(
            commandline,
            get_generic_script_command(
                commandline,
                ignore_switches=[
                    "-a",
                    "-d",
                    "--debug",
                    "--disable",
                    #
                    # Quickfix for #74, better implementations welcome!
                    # https://github.com/walles/px/issues/74
                    "-Eascii-8bit:ascii-8bit",
                    #
                    "-l",
                    "-n",
                    "-p",
                    "-s",
                    "-S",
                    "-v",
                    "--verbose",
                    "-w",
                    "-W0",
                    "-W1",
                    "-W2",
                    "--",
                ],
            ),
        )

    if command == "sudo":
        return faillog(commandline, get_sudo_command(commandline))

    if command in [
        # NOTE: This list contains binaries that are mostly used with
        # subcommands. Scripts (like brew.rb) are handled in
        # get_generic_script_command().
        #
        # "gradle" and "mvn" are not handled at all, could be added to
        # get_java_command().
        "apt-get",
        "apt",
        "cargo",
        "docker",
        "docker-compose",
        "git",
        "go",
        "npm",
        "pip",
        "pip3",
        "rustup",
    ]:
        return faillog(commandline, get_with_subcommand(commandline))

    if command == "terraform":
        return faillog(
            commandline, get_with_subcommand(commandline, ignore_switches=["-chdir"])
        )

    if command == "node":
        return faillog(
            commandline,
            get_generic_script_command(
                commandline,
                ignore_switches=[
                    "--max_old_space_size",
                    "--no-warnings",
                    "--enable-source-maps",
                ],
            ),
        )

    if command in ["bash", "sh", "zsh"]:
        return faillog(
            commandline,
            get_generic_script_command(commandline, ignore_switches=["-p"]),
        )

    if PERL_BIN.match(command):
        return faillog(commandline, get_generic_script_command(commandline))

    app_name_prefix = get_app_name_prefix(commandline)
    if is_human_friendly(command):
        # Already human friendly, prefer keeping it short
        app_name_prefix = ""

    if len(command) < 25:
        return app_name_prefix + command

    command_split = command.split(".")
    if len(command_split) > 1:
        command_suggestion = ""
        if len(command_split[-1]) > 4:
            # Pretend all the dots are a kind of path and go for the last
            # part only
            command_suggestion = command_split[-1]
        else:
            # Assume last part is a file suffix (like ".exe") and take the
            # next to last part
            command_suggestion = command_split[-2]
        if len(command_suggestion) >= 5:
            # Good enough!
            command = command_suggestion

    return app_name_prefix + command


def get_guile_command(commandline: str) -> Optional[str]:
    array = to_array(commandline)
    array = list(filter(lambda s: s, array))

    # Ignore some switches.
    IGNORE_SWITCHES = [
        "-s",
        "--listen",
        "-ds",
        "--debug",
        "--no-debug",
        "--auto-compile",
        "--fresh-auto-compile",
        "--no-auto-compile",
        "-q",
        "--r6rs",
        "--r7rs",
        "-h",
        "--help",
        "-v",
        "--version",
    ]
    IGNORE_ARGFUL_SWITCHES = [
        "-L",
        "-C",
        "-x",
        "-l",
        "-e",
    ]
    while len(array) > 1 and (
        array[1] in IGNORE_SWITCHES
        or array[1] in IGNORE_ARGFUL_SWITCHES
        or array[1].startswith("--language=")
        or array[1].startswith("--listen=")
        or array[1].startswith("--use-srfi=")
    ):
        if array[1] in IGNORE_ARGFUL_SWITCHES:
            if array[1] == "-l" and len(array) > 2:
                # That guile script will be executed first
                return array[2]
            else:
                del array[1:3]
        else:
            del array[1]

    if len(array) == 1:
        return None

    if array[1].startswith("-"):
        return None

    return os.path.basename(array[1])


def get_python_command(commandline: str) -> Optional[str]:
    """Returns None if we failed to figure out the script name"""
    array = to_array(commandline)
    array = list(filter(lambda s: s, array))

    # Ignore some switches, list inspired by 'python2.7 --help'
    IGNORE_SWITCHES = [
        "-b",
        "-bb",
        "-B",
        "-d",
        "-E",
        "-i",
        "-O",
        "-OO",
        "-R",
        "-s",
        "-S",
        "-t",
        "-tt",
        "-u",
        "-v",
        "-Werror",
        "-x",
        "-3",
    ]
    while len(array) > 1 and array[1] in IGNORE_SWITCHES:
        del array[1]

    python = os.path.basename(array[0])
    if len(array) == 1:
        return python

    if len(array) > 2:
        if array[1] == "-m" and not array[2].startswith("-"):
            return os.path.basename(array[2])

    if array[1].startswith("-"):
        return None

    if os.path.basename(array[1]) == "aws":
        # Drop the "python" part of "python aws"
        return get_aws_command(array[1:])

    return os.path.basename(array[1])


def get_aws_command(args: List[str]) -> Optional[str]:
    '''Extract "aws command subcommand" from a command line starting with "aws"'''
    result = ["aws"]
    for arg in args[1:]:
        if arg.startswith("--profile="):
            continue
        if arg.startswith("--region="):
            continue
        if arg.startswith("-"):
            break
        if os.path.sep in arg:
            break

        result.append(arg)
        if len(result) >= 4:
            # Got "aws command subcommand"
            break

    if len(result) == 4 and result[-1] != "help":
        del result[-1]

    return " ".join(result)


def get_dart_command(commandline: str) -> Optional[str]:
    """Returns None if we failed to figure out the script name"""
    array = to_array(commandline)
    dart = os.path.basename(array[0])
    if len(array) == 1:
        return dart

    for candidate in array[1:]:
        if candidate.startswith("-"):
            continue

        for _, char in enumerate(candidate):
            if not char.islower():
                # Not a subcommand, something like "/bin/hello.dart"
                return os.path.basename(candidate)

        # This was a subcommand
        return dart + " " + candidate

    # Neither subcommand nor file name found, give up
    return None


def get_sudo_command(commandline: str) -> Optional[str]:
    """Returns None if we failed to figure out the script name"""
    without_sudo = commandline[5:].strip()
    if not without_sudo:
        return "sudo"

    if without_sudo.startswith("-"):
        # Give up on options
        return None

    return "sudo " + get_command(without_sudo)


def prettify_fully_qualified_java_class(class_name: str) -> str:
    split = class_name.split(".")
    if len(split) == 1:
        return split[-1]

    if split[-1] == "Main":
        # Attempt to make "Main" class names more meaningful
        return split[-2] + "." + split[-1]

    return split[-1]


def get_java_command(commandline: str) -> Optional[str]:
    """Returns None if we failed to figure out the script name"""
    array = to_array(commandline)
    java = os.path.basename(array[0])
    if len(array) == 1:
        return java

    state = "skip next"
    for component in array:
        if not component:
            # Skip all empties
            continue

        if state == "skip next":
            if component.startswith("-"):
                # Skipping switches doesn't make sense. We're lost.
                return None
            state = "scanning"
            continue
        if state == "return next":
            if component.startswith("-"):
                # Returning switches doesn't make sense. We're lost.
                return None
            return os.path.basename(component)
        if state == "scanning":
            if component.startswith("-X"):
                continue
            if component.startswith("-D"):
                continue
            if component.startswith("-ea"):
                continue
            if component.startswith("-da"):
                continue
            if component.startswith("-agentlib:"):
                continue
            if component.startswith("-javaagent:"):
                continue
            if component.startswith("--add-modules="):
                continue
            if component == "--add-modules":
                state = "skip next"
                continue
            if component.startswith("--add-opens="):
                continue
            if component == "--add-opens":
                state = "skip next"
                continue
            if component.startswith("--add-exports="):
                continue
            if component == "--add-exports":
                state = "skip next"
                continue
            if component.startswith("--add-reads="):
                continue
            if component == "--add-reads":
                state = "skip next"
                continue
            if component.startswith("--patch-module="):
                continue
            if component == "--patch-module":
                state = "skip next"
                continue
            if component == "-server":
                continue
            if component == "-noverify":
                continue
            if component in ("-cp", "-classpath"):
                state = "skip next"
                continue
            if component == "-jar":
                state = "return next"
                continue
            if component.startswith("-"):
                # Unsupported switch, give up
                return None
            return prettify_fully_qualified_java_class(component)

        raise ValueError(f"Unhandled state <{state}> at <{component}> for: {array}")

    # We got to the end without being able to come up with a better name, give up
    return None


def get_with_subcommand(
    commandline: str, ignore_switches: Optional[List[str]] = None
) -> Optional[str]:
    array = to_array(commandline)

    if ignore_switches is None:
        ignore_switches = []
    while len(array) > 1 and array[1].split("=")[0] in ignore_switches:
        del array[1]

    command = os.path.basename(array[0])
    if len(array) == 1:
        return command

    if array[1].startswith("-"):
        # Unknown option, help!
        return None

    return f"{command} {array[1]}"


def get_generic_script_command(
    commandline: str, ignore_switches: Optional[List[str]] = None
) -> Optional[str]:
    """Returns None if we failed to figure out the script name"""
    array = to_array(commandline)

    if ignore_switches is None:
        ignore_switches = []
    while len(array) > 1 and array[1].split("=")[0] in ignore_switches:
        del array[1]

    vm = os.path.basename(array[0])
    if len(array) == 1:
        return vm

    if array[1].startswith("-"):
        # Unknown option, help!
        return None

    script = os.path.basename(array[1])
    if len(array) == 2:
        # vm + script
        return script
    if script not in ["brew.rb", "brew.sh", "yarn.js"]:
        return script

    subcommand = array[2]
    if subcommand.startswith("-"):
        # Unknown option before the subcommand
        return script

    return f"{script} {subcommand}"



import datetime
import sys
import logging
import unicodedata

import os
from . import px_poller
from . import px_process
from . import px_terminal
from . import px_sort_order
from . import px_processinfo
from . import px_process_menu
from . import px_category_bar

from typing import List
from typing import Dict
from typing import Tuple
from typing import Optional

LOG = logging.getLogger(__name__)

CMD_WHATEVER = -1
CMD_QUIT = 1
CMD_RESIZE = 2
CMD_HANDLED = 3
CMD_POLL_COMPLETE = 4

SEARCH_PROMPT_ACTIVE = px_terminal.inverse_video("Search (ENTER when done): ")
SEARCH_PROMPT_INACTIVE = "Search ('/' to edit): "
SEARCH_CURSOR = px_terminal.inverse_video(" ")

MODE_BASE = 0
MODE_SEARCH = 1

top_mode: int = MODE_BASE
search_string = ""

# Which pid were we last hovering?
last_highlighted_pid: Optional[int] = None

# Which row were we last hovering? Go for this one if
# we can't use last_pid.
last_highlighted_row: int = 0

# Has the user manually moved the highlight? If not, just stay on the top
# row even if the tow PID moves away.
highlight_has_moved: bool = False


sort_order = px_sort_order.SortOrder.CPU


def adjust_cpu_times(
    baseline: Dict[int, Tuple[datetime.datetime, float]],
    current: List[px_process.PxProcess],
) -> List[px_process.PxProcess]:
    """
    Identify processes in current that are also in baseline.

    For all matches, subtract the baseline process' CPU time usage from the
    current process' one.

    This way we get CPU times computed from when "px --top" was started, rather
    than from when each process was started.

    Neither current nor baseline are changed by this function.
    """
    pid2proc: Dict[int, px_process.PxProcess] = {}
    for proc in current:
        pid2proc[proc.pid] = proc

    for baseline_pid, baseline_times in baseline.items():
        baseline_start_time, baseline_cputime = baseline_times
        current_proc = pid2proc.get(baseline_pid)
        if current_proc is None:
            # This process is newer than the baseline
            continue

        if current_proc.start_time != baseline_start_time:
            # This PID has been reused
            continue

        if current_proc.cpu_time_seconds is None:
            # We can't subtract from None
            continue

        if baseline_cputime is None:
            # We can't subtract None
            continue

        if current_proc.cpu_time_seconds and baseline_cputime:
            current_proc.set_cpu_time_seconds(
                current_proc.cpu_time_seconds - baseline_cputime
            )

    return list(pid2proc.values())


def compute_aggregated_cpu_times(toplist: List[px_process.PxProcess]) -> None:
    """
    Compute aggregated CPU times for all processes in the toplist.

    This function modifies the toplist in place.
    """

    # First, find the root process
    root_process = toplist[0]
    while root_process.parent is not None:
        root_process = root_process.parent

    # Now, walk the process tree and compute aggregated CPU times
    def walk_tree(proc: px_process.PxProcess) -> float:
        sum = proc.cpu_time_seconds or 0
        for child in proc.children:
            sum += walk_tree(child)
        proc.set_aggregated_cpu_time_seconds(sum)
        return sum

    walk_tree(root_process)


def get_notnone_cpu_time_seconds(proc: px_process.PxProcess) -> float:
    seconds = proc.cpu_time_seconds
    if seconds is not None:
        return seconds
    return 0


def get_notnone_memory_percent(proc: px_process.PxProcess) -> float:
    percent = proc.memory_percent
    if percent is not None:
        return percent
    return 0


def sort_by_cpu_usage(
    toplist: List[px_process.PxProcess],
) -> List[px_process.PxProcess]:
    can_sort_by_time = False
    for process in toplist:
        metric = process.cpu_time_seconds
        if metric:
            can_sort_by_time = True
            break

    if can_sort_by_time:
        # There is at least one > 0 time in the process list, so sorting by time
        # will be of some use
        key = get_notnone_cpu_time_seconds
        return sorted(toplist, key=key, reverse=True)

    # No > 0 time in the process list, try CPU percentage as an approximation of
    # that. This should happen on the first iteration when ptop has just been
    # launched.
    return sorted(toplist, key=lambda process: process.cpu_percent or 0, reverse=True)


def sort_by_cpu_usage_tree(
    toplist: List[px_process.PxProcess],
) -> List[px_process.PxProcess]:
    """
    Sort the process list by aggregated CPU time, but keep the tree structure.
    """
    root_process = toplist[0]
    while root_process.parent is not None:
        root_process = root_process.parent

    def sort_children(proc: px_process.PxProcess) -> None:
        proc.children = sorted(
            proc.children,
            key=lambda child: child.aggregated_cpu_time_seconds,
            reverse=True,
        )
        for child in proc.children:
            sort_children(child)

    sort_children(root_process)

    # Now, recreate the list by flattening the tree
    flat_list = []

    def flatten(proc: px_process.PxProcess, level=0) -> None:
        proc.level = level
        flat_list.append(proc)
        for child in proc.children:
            flatten(child, level + 1)

    flatten(root_process)

    return flat_list


def get_toplist(
    baseline: Dict[int, Tuple[datetime.datetime, float]],
    current: List[px_process.PxProcess],
    sort_order=px_sort_order.SortOrder.CPU,
) -> List[px_process.PxProcess]:
    toplist = adjust_cpu_times(baseline, current)
    compute_aggregated_cpu_times(toplist)

    # Sort by interestingness last
    toplist = px_process.order_best_first(toplist)
    if sort_order == px_sort_order.SortOrder.MEMORY:
        toplist = sorted(toplist, key=get_notnone_memory_percent, reverse=True)
    elif sort_order == px_sort_order.SortOrder.CPU:
        toplist = sort_by_cpu_usage(toplist)
    elif sort_order == px_sort_order.SortOrder.AGGREGATED_CPU:
        toplist = sort_by_cpu_usage_tree(toplist)

    return toplist


def writebytes(bytestring: bytes) -> None:
    os.write(sys.stdout.fileno(), bytestring)


def get_line_to_highlight(
    toplist: List[px_process.PxProcess], max_process_count: int
) -> Optional[int]:
    global last_highlighted_pid
    global last_highlighted_row
    global highlight_has_moved

    if not toplist:
        # Toplist is empty or None
        return None

    if max_process_count <= 0:
        # No space for the highlight
        return None

    # Before the user has moved the highlight, we don't follow a particular PID
    if last_highlighted_pid is not None and highlight_has_moved:
        # Find PID line in list
        pid_line = None
        for index, process in enumerate(toplist):
            if process.pid == last_highlighted_pid:
                pid_line = index
                break
        if pid_line is not None and pid_line < max_process_count:
            last_highlighted_row = pid_line
            return pid_line

    # No PID or not found, go for the last highlighted line instead...

    # Bound highlight to toplist length and screen height
    last_highlighted_row = min(
        last_highlighted_row, max_process_count - 1, len(toplist) - 1
    )
    if last_highlighted_row < 0:
        last_highlighted_row = 0

    if last_highlighted_row > 0:
        highlight_has_moved = True

    # Stay on the top line unless the user has explicitly moved the highlight
    last_highlighted_pid = toplist[last_highlighted_row].pid

    return last_highlighted_row


def generate_header(
    filtered_processes: List[px_process.PxProcess],
    poller: px_poller.PxPoller,
    screen_columns: int,
) -> List[str]:
    assert screen_columns > 0

    sysload_line = px_terminal.bold("Sysload: ") + poller.get_loadstring()
    ramuse_line = px_terminal.bold("RAM Use: ") + poller.get_meminfo()

    if px_terminal.visual_length(sysload_line) > (screen_columns // 2 - 1):
        # Traditional header
        bar_length = screen_columns - 16
        if bar_length > 20:
            # Enough space for usable category bars. Length limit ^ picked entirely
            # arbitrarily, feel free to change it if you have a better number.
            rambar_by_program = (
                "["
                + px_category_bar.ram_by_program(bar_length, filtered_processes)
                + "]"
            )
            rambar_by_user = (
                "[" + px_category_bar.ram_by_user(bar_length, filtered_processes) + "]"
            )
        else:
            rambar_by_program = "[ ... ]"
            rambar_by_user = "[ ... ]"

        # Print header
        return [
            sysload_line,
            ramuse_line,
            "  By program: " + rambar_by_program,
            "     By user: " + rambar_by_user,
            px_terminal.bold("IO Load:      ") + poller.get_ioload_string(),
            "",
        ]

    # Make a split header
    bar_length = screen_columns // 2 - 3
    if bar_length > 20:
        # Enough space for usable category bars. Length limit ^ picked entirely
        # arbitrarily, feel free to change it if you have a better number.
        cpubar_by_program = (
            "[" + px_category_bar.cpu_by_program(bar_length, filtered_processes) + "]"
        )
        cpubar_by_user = (
            "[" + px_category_bar.cpu_by_user(bar_length, filtered_processes) + "]"
        )
        rambar_by_program = (
            "[" + px_category_bar.ram_by_program(bar_length, filtered_processes) + "]"
        )
        rambar_by_user = (
            "[" + px_category_bar.ram_by_user(bar_length, filtered_processes) + "]"
        )
    else:
        cpubar_by_program = "[ ... ]"
        cpubar_by_user = "[ ... ]"
        rambar_by_program = "[ ... ]"
        rambar_by_user = "[ ... ]"

    return [
        px_terminal.get_string_of_length(sysload_line, screen_columns // 2)
        + ramuse_line,
        px_terminal.get_string_of_length(cpubar_by_program, screen_columns // 2)
        + rambar_by_program,
        px_terminal.get_string_of_length(cpubar_by_user, screen_columns // 2)
        + rambar_by_user,
        px_terminal.bold("IO Load:      ") + poller.get_ioload_string(),
        "",
    ]


def get_screen_lines(
    toplist: List[px_process.PxProcess],
    poller: px_poller.PxPoller,
    screen_rows: int,
    screen_columns: int,
    include_footer: bool = True,
    search: Optional[str] = None,
) -> List[str]:
    """
    Note that the columns parameter is only used for layout purposes. Lines
    returned from this function will still need to be cropped before being
    printed to screen.
    """

    all_processes = toplist
    if search:
        # Note that we accept partial user name match, otherwise incrementally typing
        # a username becomes weird for the ptop user
        toplist = list(
            filter(lambda p: p.match(search, require_exact_user=False), toplist)
        )

        # Put exact search matches first. Useful for "px cat" or other short
        # search strings with tons of hits.
        search_pid = -1
        try:
            search_pid = int(search)
        except ValueError:
            pass
        toplist = sorted(
            toplist,
            key=lambda p: search in (p.command, p.username) or p.pid == search_pid,
            reverse=True,
        )

    # Hand out different amount of lines to the different sections
    footer_height = 0
    cputop_minheight = 10
    if include_footer:
        footer_height = 1

    lines = generate_header(all_processes, poller, screen_columns)

    # Create a launches section
    header_height = len(lines)
    main_area_height = screen_rows - header_height - footer_height
    launches_maxheight1 = (
        main_area_height - cputop_minheight
    )  # Give CPU top part enough space
    launches_maxheight2 = (
        main_area_height * 30
    ) // 100  # https://github.com/walles/px/issues/98
    launches_maxheight = min(launches_maxheight1, launches_maxheight2)

    launchlines: List[str] = []
    if launches_maxheight >= 3:
        launchlines = poller.get_launchcounter_lines()
        if len(launchlines) > 0:
            # Add a section header
            launchlines = [
                "",
                px_terminal.bold("Launched binaries, launch counts in (parentheses)"),
            ] + launchlines

            # Cut if we got too many lines
            launchlines = launchlines[0:launches_maxheight]

    # Compute cputop height now that we know how many launchlines we have
    cputop_height = screen_rows - header_height - len(launchlines) - footer_height

    # -2 = Section name + column headings
    max_process_count = cputop_height - 2

    # Search prompt needs one line
    max_process_count -= 1

    highlight_row = get_line_to_highlight(toplist, max_process_count)
    if top_mode == MODE_SEARCH:
        highlight_row = None

    toplist_table_lines = px_terminal.to_screen_lines(
        toplist[:max_process_count], highlight_row, sort_order
    )

    # Ensure that we cover the whole screen, even if it's higher than the
    # number of processes
    toplist_table_lines += screen_rows * [""]

    top_line = "Top processes by CPU usage"
    if sort_order == px_sort_order.SortOrder.MEMORY:
        top_line = "Top processes by memory usage"
    elif sort_order == px_sort_order.SortOrder.AGGREGATED_CPU:
        top_line = "Process tree ordered by aggregated CPU time"
    lines += [px_terminal.bold(top_line)]

    if top_mode == MODE_SEARCH:
        lines += [SEARCH_PROMPT_ACTIVE + px_terminal.bold(search or "") + SEARCH_CURSOR]
    else:
        assert top_mode == MODE_BASE
        lines += [SEARCH_PROMPT_INACTIVE + px_terminal.bold(search or "")]

    if max_process_count > 0:
        lines += toplist_table_lines[
            0 : max_process_count + 1
        ]  # +1 for the column headings

    lines += launchlines

    if include_footer:
        footer_line = (
            "  q - Quit  m - Sort order  / - Search  ↑↓ - Move  Enter - Select"
        )
        # Inverse the whole footer line
        footer_line = px_terminal.inverse_video(footer_line + 999 * " ")

        lines += [footer_line]

    return lines


def redraw(
    toplist: List[px_process.PxProcess],
    poller: px_poller.PxPoller,
    rows: int,
    columns: int,
    include_footer: bool = True,
) -> None:
    """
    Refresh display.

    The new display will be rows rows x columns columns.
    """
    lines = get_screen_lines(
        toplist, poller, rows, columns, include_footer, search=search_string
    )

    px_terminal.draw_screen_lines(lines, columns)


def handle_search_keypresses(key_sequence: px_terminal.ConsumableString) -> None:
    global search_string
    global last_highlighted_row
    global last_highlighted_pid

    # If this triggers our top_mode state machine is broken
    assert search_string is not None

    # NOTE: Uncomment to debug input characters
    # search_string = ":".join("{:02x}".format(ord(c)) for c in key_sequence._string)
    # return

    while len(key_sequence) > 0:
        if key_sequence.consume(px_terminal.KEY_BACKSPACE):
            search_string = search_string[:-1]
        elif key_sequence.consume(px_terminal.KEY_DELETE):
            search_string = search_string[:-1]
        elif key_sequence.consume(px_terminal.KEY_UPARROW):
            last_highlighted_row -= 1
            last_highlighted_pid = None
        elif key_sequence.consume(px_terminal.KEY_DOWNARROW):
            last_highlighted_row += 1
            last_highlighted_pid = None
        elif (
            key_sequence.consume(px_terminal.KEY_ENTER)
            or key_sequence._string == px_terminal.KEY_ESC
        ):
            # Exit search mode
            global top_mode
            top_mode = MODE_BASE
            return
        else:
            # Unable to consume more, give up
            break

    if len(key_sequence) == 0:
        return

    if px_terminal.KEY_ESC in key_sequence._string:
        # Some special key, unprintable, unhandled, never mind
        return

    try:
        for char in key_sequence._string:
            if unicodedata.category(char).startswith("C"):
                # Non-printable character, see:
                # http://www.unicode.org/reports/tr44/#GC_Values_Table
                return
    except TypeError:
        # Unable to type check this, let's not add it, just to be safe
        return

    search_string += key_sequence._string


def get_command(**kwargs):
    """
    Call getch() and interpret the results.
    """
    user_input = px_terminal.getch(**kwargs)
    if user_input is None:
        return None
    assert len(user_input) > 0

    global top_mode
    if top_mode == MODE_SEARCH:
        handle_search_keypresses(user_input)
        return CMD_HANDLED

    global last_highlighted_row
    global last_highlighted_pid
    global sort_order
    while len(user_input) > 0:
        if user_input.consume(px_terminal.KEY_UPARROW):
            last_highlighted_row -= 1
            last_highlighted_pid = None
        elif user_input.consume(px_terminal.KEY_DOWNARROW):
            last_highlighted_row += 1
            last_highlighted_pid = None
        elif user_input.consume(px_terminal.KEY_ENTER):
            if last_highlighted_pid is None:
                continue
            processes = px_process.get_all()
            process = px_processinfo.find_process_by_pid(
                last_highlighted_pid, processes
            )
            if not process:
                continue
            px_process_menu.PxProcessMenu(process).start()
        elif user_input.consume("/"):
            top_mode = MODE_SEARCH
            return None
        elif user_input.consume("m") or user_input.consume("M"):
            sort_order = sort_order.next()
        elif user_input.consume("q"):
            return CMD_QUIT
        elif user_input.consume(px_terminal.SIGWINCH_KEY):
            return CMD_RESIZE
        elif user_input.consume(px_poller.POLL_COMPLETE_KEY):
            return CMD_POLL_COMPLETE
        else:
            # Unable to consume anything, give up
            break

    return CMD_WHATEVER


def _top(search: str = "") -> None:
    global search_string
    search_string = search

    poller = px_poller.PxPoller(px_terminal.SIGWINCH_PIPE[1])
    poller.start()

    current = poller.get_all_processes()
    baseline = {p.pid: (p.start_time, p.cpu_time_seconds or 0.0) for p in current}

    toplist = get_toplist(baseline, current, sort_order)

    rows, columns = px_terminal.get_window_size()

    while True:
        redraw(toplist, poller, rows, columns)

        command = get_command()

        # Handle all keypresses before refreshing the display
        if command is not None:
            poller.pause_process_updates_a_bit()
            if command == CMD_QUIT:
                # The idea here is that if you terminate with "q" you still
                # probably want the heading line on screen. So just do another
                # update with somewhat fewer lines, and you'll get just that.
                redraw(toplist, poller, rows - 4, columns, include_footer=False)
                return

            if command == CMD_RESIZE:
                rows, columns = px_terminal.get_window_size()

            if command == CMD_POLL_COMPLETE:
                current = poller.get_all_processes()
                toplist = get_toplist(baseline, current, sort_order)


def top(search: str = "") -> None:
    if not sys.stdout.isatty():
        sys.stderr.write(
            'Top mode only works on TTYs, try running just "px" instead.\n'
        )
        sys.exit(1)

    with px_terminal.fullscreen_display():
        try:
            _top(search=search)
        except Exception:
            LOG.exception("Running ptop failed")

        # Make sure we actually end up on a new line
        print("")



from typing import Tuple


def bytes_to_strings(current_bytes_count: int, max_bytes_count: int) -> Tuple[str, str]:
    """
    Turn byte counts into strings like "14MB" / "16MB". Both strings will have
    the same unit, and the unit is based on the max_bytes_count value.

    Ref: https://github.com/walles/px/issues/97
    """
    KB = 1024**1
    MB = 1024**2
    GB = 1024**3
    TB = 1024**4

    unit = "B"
    divisor = 1
    if max_bytes_count < 7 * KB:
        pass
    elif max_bytes_count < 7 * MB:
        unit = "KB"
        divisor = KB
    elif max_bytes_count < 7 * GB:
        unit = "MB"
        divisor = MB
    elif max_bytes_count < 7 * TB:
        unit = "GB"
        divisor = GB
    else:
        unit = "TB"
        divisor = TB

    current_str = str(int(round(float(current_bytes_count) / divisor))) + unit
    max_str = str(int(round(float(max_bytes_count) / divisor))) + unit

    return (current_str, max_str)



from . import px_file
from . import px_process
from typing import Set
from typing import List
from typing import Dict
from typing import Text
from typing import MutableMapping
from typing import Iterable
from typing import TypeVar
from typing import Optional

T = TypeVar("T")
S = TypeVar("S")

FILE_TYPES = ["PIPE", "FIFO", "unix", "IPv4", "IPv6"]


class PeerProcess:
    def __init__(self, name: Optional[Text] = None, pid: Optional[int] = None) -> None:
        if not name:
            if pid is None:
                raise ValueError("Either pid, name or both must be set")
            name = "PID " + str(pid)
        else:
            if pid is not None:
                name += "(" + str(pid) + ")"

        self.name: Text = name
        self.pid: Optional[int] = pid

    def __repr__(self):
        return self.name

    def __hash__(self):
        return self.name.__hash__()

    def __str__(self):
        return self.name


class IpcMap:
    """
    This is a map of process->[channels], where "process" is a process we have
    IPC communication open with, and a channel is a socket or a pipe that we
    have open to that process.

    After creating an IpcMap, you can access:
    * ipc_map.network_connections: This is a list of non-IPC network connections
    * ipc_map.keys(): A set of other px_processes this process is connected to
    * ipc_map[px_process]: A set of px_files through which we're connected to the
      px_process
    """

    def __init__(
        self,
        process: px_process.PxProcess,
        files: Iterable[px_file.PxFile],
        processes: Iterable[px_process.PxProcess],
        is_root: bool,
    ) -> None:
        # On Linux, lsof reports the same open file once per thread of a
        # process. Putting the files in a set gives us each file only once.
        files = set(files)

        self._own_files = list(
            filter(lambda f: f.pid == process.pid and f.fd is not None, files)
        )

        # Only deal with IPC related files
        self.files = list(filter(lambda f: f.type in FILE_TYPES, files))

        self.process = process
        self.processes = processes
        self.ipc_files_for_process = list(
            filter(lambda f: f.pid == self.process.pid, self.files)
        )

        self._map: MutableMapping[PeerProcess, Set[px_file.PxFile]] = {}
        self._create_mapping()

        self.fds = self._create_fds(is_root)

    def _create_fds(self, is_root: bool) -> Dict[int, str]:
        """
        Describe standard FDs open by this process; the mapping is from FD number to
        FD description.

        The returned dict will always contain entries for 0, 1 and 2.

        In theory this method could easily be modified to go through all fds, not
        just the standard ones, but that can get us lots more DNS lookups, and
        take a lot of time. If you do want to try it, just drop all the "if fd
        not in [0, 1, 2]: continue"s and benchmark it on not-cached IP addresses.
        """
        fds = {}

        if not self._own_files:
            for fd in [0, 1, 2]:
                fds[fd] = "<unavailable, running px as root might help>"
            return fds

        for fd in [0, 1, 2]:
            fds[fd] = "<closed>"

        for file in self._own_files:
            if file.fd not in [0, 1, 2]:
                continue

            fds[file.fd] = str(file)

            if file.type in FILE_TYPES:
                excuse = "destination not found, try running px as root"
                if is_root:
                    excuse = "not connected"
                name: Optional[str] = file.name
                if not name:
                    name = file.device
                if name and name.startswith("->"):
                    name = name[2:]
                fds[file.fd] = f"[{file.type}] <{excuse}> ({name})"

        # Traverse network connections and update FDs as required
        for network_connection in self.network_connections:
            if network_connection.fd is None:
                continue
            if network_connection.fd not in [0, 1, 2]:
                continue
            fds[network_connection.fd] = str(network_connection)

        # Traverse our IPC structure and update FDs as required
        for target in self.keys():
            for link in self[target]:
                if link.fd is None:
                    # No FD, never mind
                    continue
                if link.fd not in [0, 1, 2]:
                    continue

                # FIXME: If this is a PIPE/FIFO leading to ourselves we should say that
                # FIXME: If this is an unconnected PIPE/FIFO, we should say that

                name = link.name
                if name and name.startswith("->"):
                    name = name[2:]
                fds[link.fd] = f"[{link.type}] -> {target} ({name})"

        return fds

    def _create_mapping(self) -> None:
        self._create_indices()

        unknown = PeerProcess(
            name="UNKNOWN destinations: Running with sudo might help find out where these go"
        )

        network_connections = set()
        for file in self.ipc_files_for_process:
            if file.type in ["FIFO", "PIPE"] and not file.fifo_id():
                # Unidentifiable FIFO, just ignore this
                continue

            other_end_pids = self._get_other_end_pids(file)
            if not other_end_pids:
                if file.type in ["IPv4", "IPv6"]:
                    # This is a remote connection
                    network_connections.add(file)
                    continue

                self.add_ipc_entry(unknown, file)
                continue

            for other_end_pid in other_end_pids:
                if other_end_pid == self.process.pid:
                    # Talking to ourselves, never mind
                    continue

                other_end_process = self._pid2process.get(other_end_pid)
                if not other_end_process:
                    other_end_process = PeerProcess(pid=other_end_pid)
                    self._pid2process[other_end_pid] = other_end_process
                self.add_ipc_entry(other_end_process, file)

        self.network_connections: Set[px_file.PxFile] = network_connections

    def _create_indices(self) -> None:
        """
        Creates indices used by _get_other_end_pids()
        """
        self._pid2process = create_pid2process(self.processes)

        self._device_to_pids: MutableMapping[str, List[int]] = {}
        self._name_to_pids: MutableMapping[str, List[int]] = {}
        self._name_to_files: MutableMapping[str, List[px_file.PxFile]] = {}
        self._device_number_to_files: MutableMapping[int, List[px_file.PxFile]] = {}
        self._fifo_id_and_access_to_pids: MutableMapping[str, List[int]] = {}
        self._local_endpoint_to_pid: MutableMapping[str, int] = {}
        for file in self.files:
            if file.device is not None:
                add_arraymapping(self._device_to_pids, file.device, file.pid)

            if file.name:
                add_arraymapping(self._name_to_pids, file.name, file.pid)
                add_arraymapping(self._name_to_files, file.name, file)

            local_endpoint, _ = file.get_endpoints()
            if local_endpoint:
                self._local_endpoint_to_pid[local_endpoint] = file.pid

            device_number = file.device_number()
            if device_number is not None:
                add_arraymapping(self._device_number_to_files, device_number, file)

            if file.access is not None and file.type == "FIFO":
                fifo_id = file.fifo_id()
                if fifo_id:
                    add_arraymapping(
                        self._fifo_id_and_access_to_pids,
                        fifo_id + file.access,
                        file.pid,
                    )

    def _get_other_end_pids(self, file: px_file.PxFile) -> Iterable[int]:
        """Locate the other end of a pipe / domain socket"""
        if file.type in ["IPv4", "IPv6"]:
            _, remote = file.get_endpoints()
            if remote is None:
                return []

            pid = self._local_endpoint_to_pid.get(remote)
            if pid:
                return [pid]
            return []

        name = file.name
        if name and name.startswith("->"):
            # With lsof 4.87 on OS X 10.11.3, pipe and socket names start with "->",
            # but their endpoint names don't. Strip initial "->" from name before
            # scanning for it.
            name = name[2:]

        file_device_with_arrow = None
        if file.device is not None:
            file_device_with_arrow = "->" + file.device

        pids: Set[int] = set()

        # The other end of the socket / pipe is encoded in the DEVICE field of
        # lsof's output ("view source" in your browser to see the conversation):
        # http://www.justskins.com/forums/lsof-find-both-endpoints-of-a-unix-socket-123037.html
        if name:
            matching_pids = self._device_to_pids.get(name)
            if matching_pids:
                pids.update(matching_pids)

        if file_device_with_arrow:
            matching_pids = self._name_to_pids.get(file_device_with_arrow)
            if matching_pids:
                pids.update(matching_pids)

        device_number = file.device_number()
        if device_number:
            matching_files = self._device_number_to_files.get(device_number)
            if not matching_files:
                matching_files = []
            for candidate in matching_files:
                if candidate.name == file.name:
                    pids.add(candidate.pid)

        fifo_id = file.fifo_id()
        if fifo_id and file.access and file.type == "FIFO":
            # On Linux, this is how we trace FIFOs
            opposing_access = {"r": "w", "w": "r"}.get(file.access)
            if opposing_access:
                name_and_opposing_access = fifo_id + opposing_access
                matching_pids = self._fifo_id_and_access_to_pids.get(
                    name_and_opposing_access
                )
                if matching_pids:
                    pids.update(matching_pids)

        return pids

    def add_ipc_entry(self, process: PeerProcess, file: px_file.PxFile) -> None:
        """
        Note that we're connected to process via file
        """
        if process not in self._map:
            self._map[process] = set()

        self._map[process].add(file)

    def keys(self) -> Iterable[PeerProcess]:
        """
        Returns a set of other px_processes this process is connected to
        """
        return self._map.keys()

    def __getitem__(self, process: PeerProcess) -> Set[px_file.PxFile]:
        """
        Returns a set of px_files through which we're connected to the px_process
        """
        return self._map.__getitem__(process)


def create_pid2process(
    processes: Iterable[px_process.PxProcess],
) -> MutableMapping[int, PeerProcess]:
    pid2process: MutableMapping[int, PeerProcess] = {}
    for process in processes:
        # Guard against duplicate PIDs
        assert process.pid not in pid2process

        pid2process[process.pid] = PeerProcess(name=process.command, pid=process.pid)

    return pid2process


def add_arraymapping(mapping: MutableMapping[S, List[T]], key: S, value: T) -> None:
    array = mapping.setdefault(key, [])
    array.append(value)



import os
import subprocess

from typing import List
from typing import Dict


ENV: Dict[str, str] = {}
for name, value in os.environ.items():
    if name == "LANG":
        continue
    if name.startswith("LC_"):
        continue
    ENV[name] = value


def run(command: List[str], check_exitcode: bool = False) -> str:
    with subprocess.Popen(
        command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=ENV
    ) as execution:
        stdout = execution.communicate()[0].decode("utf-8")

        if check_exitcode and execution.returncode != 0:
            raise subprocess.CalledProcessError(execution.returncode, command)

        return stdout



import re
import os
import random
import datetime

from px import px_file
from px import px_process
from px import px_ipc_map

import dateutil.parser

from typing import MutableMapping
from typing import Optional
from typing import List

# An example time string that can be produced by ps
TIMESTRING = "Mon Mar  7 09:33:11 2016"
TIMEZONE = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo
TIME = dateutil.parser.parse(TIMESTRING).replace(tzinfo=TIMEZONE)


def load(sample_file_name: str) -> str:
    my_dir = os.path.dirname(__file__)
    full_path = os.path.join(my_dir, sample_file_name)
    with open(full_path, encoding="utf-8") as sample_file:
        return sample_file.read()


def spaces(at_least=1, at_most=3):
    return " " * random.randint(at_least, at_most)


def local_now():
    return datetime.datetime.now().replace(tzinfo=TIMEZONE)


def create_process(
    pid=47536,
    ppid=1234,
    rss_kb=12345,
    timestring=TIMESTRING,
    uid=0,
    cpuusage="0.0",
    cputime="0:00.03",
    mempercent="0.0",
    commandline="/usr/sbin/cupsd -l",
    now=local_now(),
) -> px_process.PxProcess:
    psline = (
        spaces(at_least=0)
        + str(pid)
        + spaces()
        + str(ppid)
        + spaces()
        + str(rss_kb)
        + spaces()
        + timestring
        + spaces()
        + str(uid)
        + spaces()
        + cpuusage
        + spaces()
        + cputime
        + spaces()
        + mempercent
        + spaces()
        + commandline
    )

    return px_process.ps_line_to_process(psline, now)


def create_file(
    filetype: str,
    name: str,
    device: Optional[str],
    pid: int,
    access: Optional[str] = None,
    inode: Optional[str] = None,
    fd: Optional[int] = None,
    fdtype: Optional[str] = None,
):
    # type (...) -> px_file.PxFile

    # Remove leading [] group from name if any
    match = re.match(r"(\[[^]]*\] )?(.*)", name)
    assert match
    name = match.group(2)

    file = px_file.PxFile(pid, filetype)
    file.name = name

    file.device = device
    file.access = access
    file.inode = inode
    file.fd = fd
    file.fdtype = fdtype
    return file


def create_ipc_map(
    pid: int, all_files: List[px_file.PxFile], is_root: bool = False
) -> px_ipc_map.IpcMap:
    """Wrapper around IpcMap() so that we can test it"""
    pid2process: MutableMapping[int, px_process.PxProcess] = {}
    for file in all_files:
        if file.pid in pid2process:
            continue
        pid2process[file.pid] = create_process(pid=file.pid)
    if pid not in pid2process:
        pid2process[pid] = create_process(pid=pid)

    processes = list(pid2process.values())
    random.shuffle(processes)

    process = pid2process[pid]

    return px_ipc_map.IpcMap(process, all_files, processes, is_root)


def fake_callchain(*args: str) -> px_process.PxProcess:
    procs = []
    for arg in args:
        procs.append(create_process(commandline=arg))

    parent = None
    last_proc = None
    for proc in procs:
        proc.parent = parent
        parent = proc
        last_proc = proc

    assert last_proc is not None
    return last_proc



import os
import errno
import logging
import threading
import subprocess

from . import px_processinfo

from . import px_process
from typing import List
from typing import Optional

LOG = logging.getLogger(__name__)


def _pump_info_to_fd(with_fileno, process, processes):
    # FIXME: Type check first parameter using mypy protocols?
    # See: https://stackoverflow.com/a/56081214/473672

    # NOTE: The first parameter to this function *must* be *an object owning the fileno*,
    # not just the fileno itself. Otherwise the file will get closed when the fileno owner
    # goes out of scope in the main thread, and we'll be talking to a fileno which points
    # to who-knows-where.
    try:
        px_processinfo.print_process_info(with_fileno.fileno(), process, processes)
    except OSError as e:
        if e.errno == errno.EPIPE:
            # The user probably just exited the pager before we were done piping into it
            LOG.debug("Lost contact with pager, errno %d", e.errno)

            # FIXME: If an lsof instance was spawned by our px_processinfo.print_process_info()
            # call above (this is likely), we may want to kill that particular lsof instance.
            # It will use a bunch of CPU for some time, and we will never use its result anyway.
        else:
            LOG.error(
                "Unexpected OSError pumping process info into pager", exc_info=True
            )
    except Exception:
        # Got exc_info from: https://stackoverflow.com/a/193153/473672
        LOG.error("Failed pumping process info into pager", exc_info=True)
    finally:
        with_fileno.close()


# From: https://stackoverflow.com/a/377028/473672
def which(program: Optional[str]) -> Optional[str]:
    if not program:
        return None

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, _ = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


def to_command_line(spec: Optional[str]) -> Optional[List[str]]:
    if not spec:
        return None

    split_spec = spec.split()
    arg0 = which(split_spec[0])
    if not arg0:
        # Not found
        return None

    return [arg0] + split_spec[1:]


def launch_pager():
    env = os.environ.copy()

    # Prevent --quit-if-one-screen, we always want a pager
    env["LESS"] = ""

    # Git sets this to "FRX", but "F" means "don't page if output is too short"
    # and we always want to page, so we go for "RX" only.
    #
    # "F" might work when somebody does "px 1234", but when ptop wants to show
    # process info it w
    env["LESS"] = "RX"

    # Git does this as well
    env["LV"] = "-c"

    pager_cmd = to_command_line(env.get("PAGER", None))
    if not pager_cmd:
        # Prefer moar: https://github.com/walles/moar
        pager_cmd = to_command_line("moar")
    if not pager_cmd:
        pager_cmd = to_command_line("less")

    # FIXME: What should we do if we don't find anything to page with?
    assert pager_cmd is not None

    if pager_cmd[0].split(os.sep)[-1] == "less":
        # Prevent --quit-if-one-screen, we always want a pager
        pager_cmd = pager_cmd[0:1]

    return subprocess.Popen(pager_cmd, stdin=subprocess.PIPE, env=env)


def page_process_info(
    process: px_process.PxProcess, processes: List[px_process.PxProcess]
) -> None:
    pager = launch_pager()
    pager_stdin = pager.stdin
    assert pager_stdin is not None

    # Do this in a thread to avoid problems with pipe buffers filling up and blocking
    info_thread = threading.Thread(
        target=_pump_info_to_fd, args=(pager_stdin, process, processes)
    )

    # Terminating ptop while this is running is fine. This is deprecated since
    # Python 3.10, but we want to support older Pythons as well so let's keep it
    # this way for now.
    info_thread.setDaemon(True)

    info_thread.start()

    pagerExitcode = pager.wait()
    if pagerExitcode != 0:
        LOG.warning("Pager exited with code %d", pagerExitcode)

    # FIXME: Maybe join info_thread here as well to ensure we aren't still pumping before returning?
    # This could possibly prevent https://github.com/walles/px/issues/67



import re
import sys
import errno
import platform

from . import px_units
from . import px_terminal
from . import px_exec_util

from typing import List
from typing import Tuple
from typing import Optional


PAGE_SIZE_RE = re.compile(r"page size of ([0-9]+) bytes")

# Example input, from "sysctl vm.swapusage":
# "vm.swapusage: total = 2048.00M  used = 562.75M  free = 1485.25M  (encrypted)"
SWAPUSAGE_RE = re.compile(r".*used = ([0-9.]+)M.*")


def get_meminfo() -> str:
    total_ram_bytes, wanted_ram_bytes = _get_ram_numbers()
    percentage = (100.0 * wanted_ram_bytes) / total_ram_bytes

    percentage_string = str(int(round(percentage))) + "%"

    # "80"? I made it up.
    if percentage < 80:
        percentage_string = px_terminal.green(percentage_string)
    elif percentage < 100:
        percentage_string = px_terminal.yellow(percentage_string)
    else:
        percentage_string = px_terminal.red(percentage_string)

    current_ram, max_ram = px_units.bytes_to_strings(wanted_ram_bytes, total_ram_bytes)
    wanted_and_total_string = "".join(
        [
            px_terminal.bold(current_ram),
            " / ",
            px_terminal.bold(max_ram),
        ]
    )

    ram_text = "".join([percentage_string, "  [", wanted_and_total_string, "]"])

    return ram_text


def _get_ram_numbers() -> Tuple[int, int]:
    """
    Returns a tuple containing two numbers:
    * Total amount of RAM installed in the machine (in bytes)
    * Wanted amount of RAM by the system (in bytes)

    If wanted > total it generally implies that we're swapping.
    """
    return_me = _get_ram_numbers_from_proc()
    if return_me is not None:
        return return_me

    return_me = _get_ram_numbers_macos()
    if return_me is not None:
        return return_me

    uname = str(platform.uname())
    platform_s = uname + " Python " + sys.version

    raise IOError("Unable to get memory info " + platform_s)


def _update_from_meminfo(base: Optional[int], line: str, name: str) -> Optional[int]:
    if not line.startswith(name + ":"):
        return base

    just_the_number = line[len(name) + 1 : len(line) - 3]

    return int(just_the_number)


def _get_ram_numbers_from_proc(
    proc_meminfo: str = "/proc/meminfo",
) -> Optional[Tuple[int, int]]:
    total_kb: Optional[int] = None
    available_kb: Optional[int] = None
    free_kb: Optional[int] = None
    buffers_kb: Optional[int] = None
    cached_kb: Optional[int] = None
    swapcached_kb: Optional[int] = None
    swaptotal_kb: Optional[int] = None
    swapfree_kb: Optional[int] = None

    try:
        with open(proc_meminfo, encoding="utf-8") as f:
            for line in f:
                total_kb = _update_from_meminfo(total_kb, line, "MemTotal")
                available_kb = _update_from_meminfo(available_kb, line, "MemAvailable")
                free_kb = _update_from_meminfo(free_kb, line, "MemFree")
                buffers_kb = _update_from_meminfo(buffers_kb, line, "Buffers")
                cached_kb = _update_from_meminfo(cached_kb, line, "Cached")
                swapcached_kb = _update_from_meminfo(swapcached_kb, line, "SwapCached")
                swaptotal_kb = _update_from_meminfo(swaptotal_kb, line, "SwapTotal")
                swapfree_kb = _update_from_meminfo(swapfree_kb, line, "SwapFree")
    except (IOError, OSError) as e:
        if e.errno == errno.ENOENT:
            # /proc/meminfo not found, we're probably not on Linux
            return None

        raise

    if total_kb is None:
        return None
    if swaptotal_kb is None:
        return None
    if swapfree_kb is None:
        return None
    swapused_kb = swaptotal_kb - swapfree_kb

    if available_kb is not None:
        # Use MemAvailable if we can:
        # https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/commit/?id=34e431b0ae398fc54ea69ff85ec700722c9da773

        ramused_kb = total_kb - available_kb

        return (total_kb * 1024, (swapused_kb + ramused_kb) * 1024)

    if free_kb is None:
        return None
    if buffers_kb is None:
        return None
    if cached_kb is None:
        return None
    if swapcached_kb is None:
        return None

    ramused_kb = total_kb - (free_kb + buffers_kb + cached_kb + swapcached_kb)

    return (total_kb * 1024, (swapused_kb + ramused_kb) * 1024)


def _get_ram_numbers_macos() -> Optional[Tuple[int, int]]:
    vm_stat_lines = _get_vmstat_output_lines()
    if vm_stat_lines is None:
        return None

    vmstat_ram = _get_ram_numbers_from_vm_stat_output(vm_stat_lines)
    if vmstat_ram is None:
        return None

    total_ram_bytes, used_ram_bytes = vmstat_ram

    used_swap_bytes = _get_used_swap_bytes_sysctl()

    return (total_ram_bytes, used_ram_bytes + used_swap_bytes)


def _get_vmstat_output_lines() -> Optional[List[str]]:
    try:
        return px_exec_util.run(["vm_stat"]).splitlines()
    except (IOError, OSError) as e:
        if e.errno == errno.ENOENT:
            # vm_stat not found, we're probably not on OSX
            return None

        raise


def _get_used_swap_bytes_sysctl() -> int:
    stdout = px_exec_util.run(["sysctl", "vm.swapusage"], check_exitcode=True)
    match = SWAPUSAGE_RE.match(stdout.strip())
    if not match:
        raise IOError("No swap usage in 'sysctl vm.swapusage' output: " + stdout)

    return int(float(match.group(1)) * 1024 * 1024)


def _update_if_prefix(base: Optional[int], line: str, prefix: str) -> Optional[int]:
    if not line.startswith(prefix):
        return base

    no_ending_dot = line.rstrip(".")

    return int(no_ending_dot[len(prefix) :])


def _get_ram_numbers_from_vm_stat_output(
    vm_stat_lines: List[str],
) -> Optional[Tuple[int, int]]:
    # List based on https://apple.stackexchange.com/a/196925/182882
    page_size_bytes = None
    pages_free = None
    pages_active = None
    pages_inactive = None
    pages_speculative = None
    pages_wired = None
    pages_anonymous = None
    pages_purgeable = None
    pages_compressed = None  # "Pages occupied by compressor"
    pages_uncompressed = None  # "Pages stored in compressor"

    for line in vm_stat_lines:
        page_size_match = PAGE_SIZE_RE.search(line)
        if page_size_match:
            page_size_bytes = int(page_size_match.group(1))
            continue

        pages_free = _update_if_prefix(pages_free, line, "Pages free:")
        pages_active = _update_if_prefix(pages_active, line, "Pages active:")
        pages_inactive = _update_if_prefix(pages_inactive, line, "Pages inactive:")
        pages_speculative = _update_if_prefix(
            pages_speculative, line, "Pages speculative:"
        )
        pages_wired = _update_if_prefix(pages_wired, line, "Pages wired down:")
        pages_anonymous = _update_if_prefix(pages_anonymous, line, "Anonymous pages:")
        pages_purgeable = _update_if_prefix(pages_purgeable, line, "Pages purgeable:")
        pages_compressed = _update_if_prefix(
            pages_compressed, line, "Pages occupied by compressor:"
        )
        pages_uncompressed = _update_if_prefix(
            pages_uncompressed, line, "Pages stored in compressor:"
        )

    if page_size_bytes is None:
        return None
    if pages_free is None:
        return None
    if pages_active is None:
        return None
    if pages_inactive is None:
        return None
    if pages_speculative is None:
        return None
    if pages_wired is None:
        return None
    if pages_anonymous is None:
        return None
    if pages_purgeable is None:
        return None
    if pages_compressed is None:
        return None
    if pages_uncompressed is None:
        return None

    # In experiments, this has added up well to the amount of physical RAM in my machine
    total_ram_pages = (
        pages_free
        + pages_active
        + pages_inactive
        + pages_speculative
        + pages_wired
        + pages_compressed
    )

    # This matches what the Activity Monitor shows in macOS 10.15.6
    #
    # For anonymous - purgeable: https://stackoverflow.com/a/36721309/473672
    #
    # FIXME: We want to add swapped out pages to this as well, since those also
    # represent a want for pages.
    wanted_ram_pages = (
        pages_anonymous - pages_purgeable + pages_wired + pages_compressed
    )

    total_ram_bytes = total_ram_pages * page_size_bytes
    wanted_ram_bytes = wanted_ram_pages * page_size_bytes

    return (total_ram_bytes, wanted_ram_bytes)



"""
Interactive menu for killing or infoing a process.

Invoked from px_top.py.
"""

import os
import time
import errno
import subprocess

from . import px_pager
from . import px_process
from . import px_terminal
from . import px_processinfo

from typing import Callable

# Constants signal.SIGXXX are enums in Python 3. But we want the numbers (to
# pass as an argument to /bin/kill), so we make our own int constants.
SIGTERM = 15
SIGKILL = 9

KILL_TIMEOUT_SECONDS = 5


def get_header_line(process: px_process.PxProcess) -> str:
    header_line = "Process: "
    header_line += str(process.pid) + " " + process.command
    header_line = px_terminal.bold(header_line)
    return header_line


def kill(process: px_process.PxProcess, signo: int) -> bool:
    """
    Signal a process.

    Returns True if the signal was delivered, False otherwise (not allowed).
    """
    try:
        os.kill(process.pid, signo)
    except (IOError, OSError) as e:
        if e.errno not in [errno.EPERM, errno.EACCES]:
            raise e

        return False

    return True


def sudo_kill(process: px_process.PxProcess, signo: int) -> bool:
    """
    Signal a process as root.

    Returns True if the signal was delivered, False otherwise.
    """
    with px_terminal.normal_display():
        print(px_terminal.CLEAR_SCREEN)

        # Print process screen heading followed by an empty line
        _, columns = px_terminal.get_window_size()

        print(px_terminal.crop_ansi_string_at_length(get_header_line(process), columns))
        print("")

        # Print "sudo kill 1234"
        command = ["sudo", "kill"]
        if signo != SIGTERM:
            command += ["-" + str(signo)]
        command += [str(process.pid)]

        print("$ " + " ".join(command))

        # Invoke "sudo kill 1234"
        returncode = subprocess.call(command)
        if returncode == 0:
            return True

        # Give user time to peruse any error message
        time.sleep(1.5)
        return False


class PxProcessMenu:
    # NOTE: Must match number constants in execute_menu_entry()
    MENU_ENTRIES = [
        "Show info",
        "Kill process",
        "Kill process as root",
        "Back to process listing",
    ]

    def __init__(self, process: px_process.PxProcess) -> None:
        self.process = process
        self.done = False

        # Shown to user, status of last operation
        self.status = ""

        # Index into MENU_ENTRIES
        self.active_entry = 0

    def refresh_display(self) -> None:
        _, columns = px_terminal.get_window_size()

        lines = []

        lines += [get_header_line(self.process)]
        lines += [""]

        lines += [
            px_terminal.bold("Arrow keys")
            + " move up and down, "
            + px_terminal.bold("RETURN")
            + " selects, "
            + px_terminal.bold("ESC")
            + " to go back."
        ]
        lines += [""]

        last_entry_no = len(self.MENU_ENTRIES) - 1
        for entry_no, text in enumerate(self.MENU_ENTRIES):
            prefix = "    "
            arrow = "⇵"
            if entry_no == 0:
                arrow = "↓"
            elif entry_no == last_entry_no:
                arrow = "↑"
            if entry_no == self.active_entry:
                prefix = arrow + " ->"
                text = px_terminal.inverse_video(text)

            lines += [prefix + text]

        if self.status:
            lines += ["", "Status: " + px_terminal.bold(self.status)]

        px_terminal.draw_screen_lines(lines, columns)

    def await_and_handle_user_input(self) -> None:
        incoming = px_terminal.getch()
        if incoming is None:
            return
        assert len(incoming) > 0

        self.status = ""
        while len(incoming) > 0:
            if incoming.consume(px_terminal.KEY_UPARROW):
                self.active_entry -= 1
                if self.active_entry < 0:
                    self.active_entry = 0
            elif incoming.consume(px_terminal.KEY_DOWNARROW):
                self.active_entry += 1
                if self.active_entry >= len(self.MENU_ENTRIES):
                    self.active_entry = len(self.MENU_ENTRIES) - 1
            elif incoming.consume(px_terminal.KEY_ENTER):
                self.execute_menu_entry()
            elif incoming.consume("q"):
                self.done = True
                return
            elif incoming.consume(px_terminal.SIGWINCH_KEY):
                # After we return the screen will be refreshed anyway,
                # no need to do anything here.
                continue
            elif incoming.consume(px_terminal.KEY_ESC):
                self.done = True
                return
            else:
                # Unable to consume, give up
                break

    def start(self) -> None:
        """
        Process menu main loop
        """
        while (not self.done) and (self.process.is_alive()):
            self.refresh_display()
            self.await_and_handle_user_input()

    def page_process_info(self) -> None:
        """
        Display process info in a pager.
        """
        processes = px_process.get_all()
        process = px_processinfo.find_process_by_pid(self.process.pid, processes)
        if not process:
            # Process not available, never mind
            return

        with px_terminal.normal_display():
            px_pager.page_process_info(process, processes)

    def await_death(self, message):
        # type(str) -> None
        """
        Wait KILL_TIMEOUT_SECONDS for process to die.

        Returns after either the process dies or we run out of time,
        whichever comes first.
        """
        t0 = time.time()
        while (time.time() - t0) < KILL_TIMEOUT_SECONDS:
            if not self.process.is_alive():
                return

            dt_s = time.time() - t0
            countdown_s = KILL_TIMEOUT_SECONDS - dt_s
            if countdown_s <= 0:
                return
            self.status = f"{countdown_s:.1f}s {message}"
            self.refresh_display()

            time.sleep(0.1)

    def kill_process(
        self, signal_process: Callable[[px_process.PxProcess, int], bool]
    ) -> None:
        """
        Send first SIGTERM then SIGKILL to a process.

        Wait KILL_TIMEOUT_SECONDS secods in between to give it a chance to go away.
        """

        # Please go away
        if not signal_process(self.process, SIGTERM):
            self.status = (
                "Not allowed to kill <" + self.process.command + ">, try again as root!"
            )
            return
        self.await_death(
            f"Waiting for {self.process.command} to shut down after SIGTERM"
        )
        if not self.process.is_alive():
            return

        # Die!!
        assert signal_process(self.process, SIGKILL)
        self.await_death(
            f"Waiting for {self.process.command} to shut down after kill -9"
        )
        if not self.process.is_alive():
            return

        self.status = "<" + self.process.command + "> did not die!"
        return

    def execute_menu_entry(self):
        # NOTE: Constants here must match lines in self.MENU_ENTRIES
        # at the top of this file
        if self.active_entry == 0:
            self.page_process_info()
        elif self.active_entry == 1:
            self.kill_process(kill)
        elif self.active_entry == 2:
            self.kill_process(sudo_kill)
        elif self.active_entry == 3:
            self.done = True



import os
import time
import threading

from . import px_load
from . import px_ioload
from . import px_meminfo
from . import px_process
from . import px_launchcounter

from typing import Optional
from typing import List


# We'll report poll done as this key having been pressed.
#
# NOTE: This must be detected as non-printable by handle_search_keypress().
POLL_COMPLETE_KEY = "\x01"

# Key repeat speed is about one every 30+ms, and this pause needs to be longer
# than that for the pause to be useful while scrolling.
SHORT_PAUSE_SECONDS = 0.1


class PxPoller:
    def __init__(self, poll_complete_notification_fd: Optional[int] = None) -> None:
        """
        After a poll is done and there is new data, a POLL_COMPLETE_KEY will be
        written to the poll_complete_notification_fd file descriptor.
        """
        self.thread: Optional[threading.Thread] = None

        self.poll_complete_notification_fd = poll_complete_notification_fd

        self.lock = threading.Lock()

        self._ioload = px_ioload.PxIoLoad()
        self._ioload_string = "None"

        self._loadstring = "None"

        self._meminfo = "None"

        self._all_processes: List[px_process.PxProcess] = []

        self._launchcounter = px_launchcounter.Launchcounter()
        self._launchcounter_screen_lines: List[str] = []

        # No process polling until this timestamp, timestamp from time.time()
        self._pause_process_updates_until = 0.0

        # Ensure we have current data already at the start
        self.poll_once()

    def pause_process_updates_a_bit(self):
        with self.lock:
            self._pause_process_updates_until = time.time() + SHORT_PAUSE_SECONDS

    def start(self):
        assert not self.thread

        self.thread = threading.Thread(name="Poller", target=self.poller)
        self.thread.daemon = True
        self.thread.start()

    def poll_once(self):
        with self.lock:
            if time.time() < self._pause_process_updates_until:
                return

        # Poll processes
        all_processes = px_process.get_all()
        with self.lock:
            self._all_processes = all_processes

        # Keep a launchcounter rendering up to date
        self._launchcounter.update(all_processes)
        launchcounter_screen_lines = self._launchcounter.get_screen_lines()
        with self.lock:
            self._launchcounter_screen_lines = launchcounter_screen_lines

        # Poll memory
        meminfo = px_meminfo.get_meminfo()
        with self.lock:
            self._meminfo = meminfo

        # Poll system load
        load = px_load.get_load_values()
        loadstring = px_load.get_load_string(load)
        with self.lock:
            self._loadstring = loadstring

        # Poll IO
        self._ioload.update()
        ioload_string = self._ioload.get_load_string()
        with self.lock:
            self._ioload_string = ioload_string

        # Notify fd that we have new data
        if self.poll_complete_notification_fd is not None:
            os.write(
                self.poll_complete_notification_fd, POLL_COMPLETE_KEY.encode("utf-8")
            )

    def poller(self):
        while True:
            with self.lock:
                sleeptime = self._pause_process_updates_until - time.time()
            if sleeptime < 0.0:
                sleeptime = 1.0
            time.sleep(sleeptime)
            self.poll_once()

    def get_all_processes(self) -> List[px_process.PxProcess]:
        with self.lock:
            return self._all_processes

    def get_ioload_string(self) -> str:
        with self.lock:
            return self._ioload_string

    def get_launchcounter_lines(self) -> List[str]:
        with self.lock:
            return self._launchcounter_screen_lines

    def get_meminfo(self) -> str:
        with self.lock:
            return self._meminfo

    def get_loadstring(self) -> str:
        with self.lock:
            return self._loadstring



import operator

from px import px_terminal

from . import px_process

from typing import Callable, List, Optional
from typing import Dict
from typing import Tuple


def cluster_processes(
    all_processes: List[px_process.PxProcess],
    get_category: Callable[[px_process.PxProcess], str],
    get_value: Callable[[px_process.PxProcess], Optional[float]],
) -> List[Tuple[str, float]]:
    """
    Group processes by category and value, summing up the values in each group.
    """

    names_to_kilobytes: Dict[str, float] = {}
    for process in all_processes:
        category = get_category(process)
        value = get_value(process)
        if value is None:
            continue

        total = names_to_kilobytes.get(category, 0)
        names_to_kilobytes[category] = total + value

    return sorted(names_to_kilobytes.items(), key=operator.itemgetter(1), reverse=True)


def render_bar(bar_length: int, names_and_numbers: List[Tuple[str, float]]) -> str:
    """
    You probably want to use x_by_y() functions at the end of this file instead,
    this is just an internal utility function.
    """

    total = 0.0
    for category in names_and_numbers:
        total += category[1]
    if total == 0:
        return ""

    bar = ""
    bar_chars = 0
    chunk_number = -1
    should_alternate = False
    while bar_chars < bar_length:
        chunk_number += 1

        if chunk_number >= len(names_and_numbers):
            # Ran out of names_and_numbers, just start alternating
            should_alternate = True

        if not should_alternate:
            # Show some real data
            name_and_number = names_and_numbers[chunk_number]
            name = name_and_number[0]
            number = name_and_number[1]

            chars = int(round(bar_length * number * 1.0 / total))
            if chars > 1:
                add_to_bar = px_terminal.get_string_of_length(" " + name, chars)
            else:
                should_alternate = True

        if should_alternate:
            add_to_bar = " "
        add_to_bar_chars = len(add_to_bar)

        if chunk_number == 0:
            # First red
            add_to_bar = px_terminal.red(add_to_bar)
        elif chunk_number == 1:
            # Second yellow
            add_to_bar = px_terminal.yellow(add_to_bar)
        elif chunk_number % 2 == 1:
            # Then alternating between inverse video and blue
            add_to_bar = px_terminal.inverse_video(add_to_bar)
        else:
            # chunk_number % 2 == 0
            add_to_bar = px_terminal.blue(add_to_bar)

        assert len(add_to_bar) > 0
        bar += add_to_bar
        bar_chars += add_to_bar_chars

    return bar


def ram_by_program(length: int, all_processes: List[px_process.PxProcess]) -> str:
    return render_bar(
        length,
        cluster_processes(
            all_processes,
            lambda process: process.command,
            lambda process: process.rss_kb,
        ),
    )


def ram_by_user(length: int, all_processes: List[px_process.PxProcess]) -> str:
    return render_bar(
        length,
        cluster_processes(
            all_processes,
            lambda process: process.username,
            lambda process: process.rss_kb,
        ),
    )


def create_cpu_getter(
    all_processes: List[px_process.PxProcess],
) -> Callable[[px_process.PxProcess], Optional[float]]:
    """
    Getter for cpu_time_seconds if there are any, otherwise for cpu_percent (of
    which we know there are a bunch).
    """
    for process in all_processes:
        if process.cpu_time_seconds is None:
            continue
        if process.cpu_time_seconds > 0:
            return lambda p: p.cpu_time_seconds

    return lambda p: p.cpu_percent


def cpu_by_program(length: int, all_processes: List[px_process.PxProcess]) -> str:
    return render_bar(
        length,
        cluster_processes(
            all_processes,
            lambda process: process.command,
            create_cpu_getter(all_processes),
        ),
    )


def cpu_by_user(length: int, all_processes: List[px_process.PxProcess]) -> str:
    return render_bar(
        length,
        cluster_processes(
            all_processes,
            lambda process: process.username,
            create_cpu_getter(all_processes),
        ),
    )



import enum


class SortOrder(enum.Enum):
    CPU = 1
    MEMORY = 2
    AGGREGATED_CPU = 3

    def next(self):
        if self == SortOrder.CPU:
            return SortOrder.MEMORY
        if self == SortOrder.MEMORY:
            return SortOrder.AGGREGATED_CPU
        return SortOrder.CPU



from . import px_process
from . import px_file
from typing import List
from typing import Dict
from typing import Optional


def _strip_leading_dash(process: px_process.PxProcess) -> str:
    key = process.command
    if key.startswith("-"):
        key = key[1:]
    return key


class PxCwdFriends:
    def __init__(
        self,
        process: px_process.PxProcess,
        all_processes: List[px_process.PxProcess],
        all_files: List[px_file.PxFile],
    ) -> None:
        pid_to_process: Dict[int, px_process.PxProcess] = {}
        for p in all_processes:
            pid_to_process[p.pid] = p

        # Cwd can be None if lsof and process listing are out of sync
        self.cwd: Optional[str] = None

        cwd_to_processes: Dict[str, List[px_process.PxProcess]] = {}
        for current_file in all_files:
            if not current_file.name:
                continue

            if current_file.fdtype != "cwd":
                continue

            if current_file.pid == process.pid:
                self.cwd = current_file.name

            if current_file.name == "/":
                # This is too common, no point in doing this one
                continue

            file_processes = cwd_to_processes.get(current_file.name)
            if file_processes is None:
                file_processes = []

            file_process = pid_to_process.get(current_file.pid)
            if file_process is not None:
                # Process could be None because there's no way for uss to get a process listing
                # a file listing that can be guaranteed to be in sync
                file_processes.append(file_process)
                cwd_to_processes[current_file.name] = file_processes

        if self.cwd is None:
            friends: List[px_process.PxProcess] = []
        elif self.cwd in cwd_to_processes:
            friends = cwd_to_processes[self.cwd]
        else:
            friends = []

        if process in friends:
            friends.remove(process)

        # Sort primarily by command and secondarily by PID
        friends = sorted(friends, key=lambda friend: friend.pid)
        friends = sorted(friends, key=_strip_leading_dash)
        self.friends: List[px_process.PxProcess] = friends
